import numpy as np

class StreamPower(object):
    """
    UNDER DEVELOPMENT. DO NOT ATTEMPT TO USE YET. DEJH Sept 2013.
    """
    
    def __init__(self, grid, data):
        self.m = 0.5
        self.n = 1.
        self.K = 1.e-5
        #This will us the MPD once finalized
        
    def self.initialize(self, grid, data):
        self.__init__(grid, data)
        
    def stream_power_erosion(self, grid, data):
        grads_at_links = grid.calculate_gradients_at_active_links(data.elev) #Make this!


#! /usr/bin/env python

from landlab import RasterModelGrid
from landlab.components.craters.component import CratersComponent


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--impact-count', type=int, default=128,
                        help='Number of impacts')
    parser.add_argument('--grid-size', type=int, default=128,
                        help='Number of grid rows and columns')
    parser.add_argument('--grid-spacing', type=float, default=.025,
                        help='Spacing between grid rows and columns')

    seed_parser = parser.add_mutually_exclusive_group()
    seed_parser.add_argument('--with-seed', type=int, default=1973,
                             help='Seed for random number generator')
    seed_parser.add_argument('--without-seed', action='store_true',
                             default=False,
                             help='Seed for random number generator')


    args = parser.parse_args()

    shape = (args.grid_size, args.grid_size)
    spacing = (args.grid_spacing, args.grid_spacing)
    impact_count = args.impact_count
    if args.without_seed:
        seed = None
    else:
        seed = args.with_seed

    grid = RasterModelGrid(shape[0], shape[1], spacing[0])

    craters_comp = CratersComponent(grid, seed=seed)
    for _ in xrange(impact_count):
        craters_comp.update()

    craters_comp.imshow('planet_surface__elevation', grid_units=('km', 'km'),
                        symmetric_cbar=True)


if __name__ == "__main__":
    main()



#! /usr/env/python
"""
Python implementation of ModelGrid, a class used to
create and manage grids for 2D numerical models.

First version GT, July 2010
Last modified August 2013
"""

import numpy
import warnings

from landlab.utils import count_repeated_values
from landlab.field import ModelDataFields

BAD_INDEX_VALUE = numpy.iinfo(numpy.int).max

# Map names grid elements to the ModelGrid attribute that contains the count
# of that element in the grid.
_ARRAY_LENGTH_ATTRIBUTES = {
    'node': 'num_nodes',
    'cell': 'num_active_cells',
    'link': 'num_active_links',
    'face': 'num_faces',
}

# Define the boundary-type codes
INTERIOR_NODE = 0
FIXED_VALUE_BOUNDARY = 1
FIXED_GRADIENT_BOUNDARY = 2
TRACKS_CELL_BOUNDARY = 3
INACTIVE_BOUNDARY = 4

BOUNDARY_STATUS_FLAGS_LIST = [
    FIXED_VALUE_BOUNDARY,
    FIXED_GRADIENT_BOUNDARY,
    TRACKS_CELL_BOUNDARY,
    INACTIVE_BOUNDARY,
]
BOUNDARY_STATUS_FLAGS = set(BOUNDARY_STATUS_FLAGS_LIST)


class Error(Exception):
    """
    Base class for exceptions from this module.
    """
    pass


def _sort_points_into_quadrants(x, y, nodes):
    """
    Divide points with locations given in the *x*, and *y* arrays into north,
    south, east, and west quadrants. Returns nodes contained in quadrants
    (west, east, north, south).

    >>> x = numpy.array([0, 1, 0, -1])
    >>> y = numpy.array([1, 0, -1, 0])
    >>> nodes = numpy.array([1, 2, 3, 4])
    >>> _sort_points_into_quadrants(x, y, nodes)
    (array([4]), array([2]), array([1]), array([3]))
    """
    above_x_axis = y > 0
    right_of_y_axis = x > 0
    closer_to_y_axis = numpy.abs(y) >= numpy.abs(x)

    north_nodes = nodes[above_x_axis & closer_to_y_axis]
    south_nodes = nodes[(~ above_x_axis) & closer_to_y_axis]
    east_nodes = nodes[right_of_y_axis & (~ closer_to_y_axis)]
    west_nodes = nodes[(~ right_of_y_axis) & (~ closer_to_y_axis)]

    return (west_nodes, east_nodes, north_nodes, south_nodes)


class ModelGrid(ModelDataFields):
    """
    Base class for creating and manipulating 2D structured or
    unstructured grids for numerical models.
    
    The idea is to have at least two inherited
    classes, RasterModelGrid and DelaunayModelGrid, that can create and
    manage grids. To this might be added a GenericModelGrid, which would
    be an unstructured polygonal grid that doesn't necessarily obey or
    understand the Delaunay triangulation, but rather simply accepts
    an input grid from the user. Also a HexModelGrid for hexagonal.
    """

    # Debugging flags (if True, activates some output statements)
    DEBUG_VERBOSE = False
    DEBUG_TRACK_METHODS = False

    #-------------------------------------------------------------------
    def __init__(self):
        for centering in ['node', 'cell', 'link', 'face', ]:
            try:
                self.add_group(centering, self._array_length(centering))
            except AttributeError:
                pass
    
    #-------------------------------------------------------------------
    def initialize( self ):
    
        pass
        
    def get_node_status(self):
        """
        Returns an array of node boundary-status codes.
        """
        return self.node_status

    def create_cell_dvector( self ):
        """
        Returns a vector of floating point numbers the same length as 
        the number of cells.
        """
    
        return numpy.zeros( self.num_cells )

    def create_active_cell_dvector( self ):
        """
        Returns a vector of floating point numbers the same length as 
        the number of active cells.
        """
    
        return numpy.zeros( self.num_active_cells )

    def create_node_dvector( self ):
        """
        Returns a vector of floating point numbers the same length as 
        the number of nodes.
        """
    
        return numpy.zeros( self.num_nodes )
        
    def create_link_dvector(self):
        """
        Returns a 1D numpy array of floating point numbers the same length as 
        the number of links.
        """
        
        return numpy.zeros(self.num_links)

    def create_active_link_dvector(self):
        """
        Returns a 1D numpy array of floating point numbers the same length as 
        the number of active links.
        """
        
        return numpy.zeros(self.num_active_links)

    def create_face_dvector( self ):
        """
        Returns a vector of floating point numbers the same length as 
        the number of interior faces.
        """
    
        return numpy.zeros( self.num_faces )

    def _array_length(self, centering):
        try:
            return getattr(self, _ARRAY_LENGTH_ATTRIBUTES[centering])
        except KeyError:
            raise TypeError('centering value not understood')

    def zeros(self, **kwds):
        """
        Returns a numpy array of zeros that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        return numpy.zeros(self._array_length(centering), **kwds)

    def empty(self, **kwds):
        """
        Returns a numpy array of uninitialized values that is the same length
        as the number of nodes in the grid. Use the *centering* keyword to
        return an array for other elements of the grid. *centering* is a
        string that is one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        return numpy.empty(self._array_length(centering), **kwds)

    def ones(self, **kwds):
        """
        Returns a numpy array of ones that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        return numpy.ones(self._array_length(centering), **kwds)

    def set_fixed_value_boundaries(self, node_ids):
        """
        Assignes FIXED_VALUE_BOUNDARY status to specified nodes.
        """
        self.node_status[node_ids] = FIXED_VALUE_BOUNDARY
        self.reset_list_of_active_links()

    def calculate_gradients_at_active_links(self, s, gradient=None):
        """
        Calculates the gradient in quantity s at each active link in the grid.
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.calculate_gradients_at_active_links'
        
        if gradient==None:
            gradient = numpy.zeros(self.num_active_links)
            
        assert (len(gradient)==self.num_active_links), \
                "len(gradient)!=num_active_links"
                
        active_link_id = 0
        for link_id in self.active_links:
            gradient[active_link_id] = (s[self.link_tonode[link_id]]
                                        -s[self.link_fromnode[link_id]]) / \
                                        self.link_length[link_id]
            active_link_id += 1
        
        return gradient
        
    def calculate_flux_divergence_at_active_cells(self, active_link_flux, 
                                                  net_unit_flux=False):
        """
        Given an array of fluxes along links, computes the net total flux
        within each cell, divides by cell area, and stores the result in
        net_unit_flux. 
        
        The input active_link_flux should be flux of
        something (e.g., mass, momentum, energy) per unit face width, positive
        if flowing in the same direction as its link, and negative otherwise.
        There should be one value per active link. Returns an array of net
        total flux per unit area, one value per active cell (creates this
        array if it is not given as an argument).
          By convention, divergence is positive for net outflow, and 
        negative for net outflow. That's why we *add* outgoing flux and
        *subtract* incoming flux. This makes net_unit_flux have the same sign
        and dimensions as a typical divergence term in a conservation equation.

        In general, for a polygonal cell with N sides of lengths
        Li and with surface area A, the net influx divided by cell
        area would be:
            .. math::
                {Q_{net} \over A} = {1 \over A} \sum{q_i L_i}

        .. note::
            The net flux is defined as positive outward, negative
            inward. In a diffusion problem, for example, one would use:
                .. math::
                    {du \over dt} = \\text{source} - \\text{fd}
            where fd is "flux divergence".
            
        .. todo::
            This needs to be re-implemented with the faster algorithm used in
            the RasterModelGrid version.
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.calculate_flux_divergence_at_active_cells'
            
        assert (len(active_link_flux)==self.num_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux==False:
            net_unit_flux = numpy.zeros(self.num_active_cells)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux))==self.num_active_cells
        
        # For each active link, add up the flux out of the "from" cell and 
        # into the "to" cell.
        active_link_id = 0
        for link_id in self.active_links:
            from_cell = self.node_activecell[self.link_fromnode[link_id]]
            to_cell = self.node_activecell[self.link_tonode[link_id]]
            total_flux = active_link_flux[active_link_id] * \
                         self.face_width[self.link_face[link_id]]
            #print('Flux '+str(total_flux)+' from '+str(from_cell) \
            #      +' to '+str(to_cell)+' along link '+str(link_id))
            if from_cell != BAD_INDEX_VALUE:
                net_unit_flux[from_cell] += total_flux
                #print('cell '+str(from_cell)+' net='+str(net_unit_flux[from_cell]))
            if to_cell != BAD_INDEX_VALUE:
                net_unit_flux[to_cell] -= total_flux
                #print('cell '+str(to_cell)+' net='+str(net_unit_flux[to_cell]))
            active_link_id += 1
        
        # Divide by cell area
        net_unit_flux = net_unit_flux / self.active_cell_areas
        
        return net_unit_flux
        
    def calculate_flux_divergence_at_nodes(self, active_link_flux, 
                                           net_unit_flux=False):
        """
        Same as calculate_flux_divergence_at_active_cells, but works with and
        returns a list of net unit fluxes that corresponds to all nodes, rather
        than just active cells. 
        
        Note that we don't compute net unit fluxes at
        boundary nodes (which don't have active cells associated with them, and 
        often don't have cells of any kind, because they are on the perimeter), 
        but simply return zeros for these entries. The advantage is that the 
        caller can work with node-based arrays instead of active-cell-based 
        arrays.

        .. todo::
            This needs to be re-implemented with the faster algorithm used in
            the RasterModelGrid version.
        """
        
        #print 'cfdn here'
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.calculate_flux_divergence_at_nodes'
            
        assert (len(active_link_flux)==self.num_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux==False:
            net_unit_flux = numpy.zeros(self.num_nodes)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux))==self.num_nodes
        
        # For each active link, add up the flux out of the "from" cell and 
        # into the "to" cell.
        active_link_id = 0
        for link_id in self.active_links:
            from_node = self.link_fromnode[link_id]
            from_cell = self.node_activecell[from_node]
            to_node = self.link_tonode[link_id]
            to_cell = self.node_activecell[to_node]
            total_flux = active_link_flux[active_link_id] * \
                         self.face_width[self.link_face[link_id]]
            #print('Flux '+str(total_flux)+' from '+str(from_node) \
            #      +' to '+str(to_node)+' along link '+str(link_id))
            if from_cell != BAD_INDEX_VALUE:
                net_unit_flux[from_node] += total_flux / \
                                            self.active_cell_areas[from_cell]
                #if from_node==46:
                #    print('node '+str(from_node)+' net='+str(net_unit_flux[from_node]))
                #    print 'fw=', self.face_width[self.link_face[link_id]]
            if to_cell != BAD_INDEX_VALUE:
                net_unit_flux[to_node] -= total_flux / \
                                          self.active_cell_areas[to_cell]
                #if to_node==46:
                #    print('node '+str(to_node)+' net='+str(net_unit_flux[to_node]))
                #    print 'fw=', self.face_width[self.link_face[link_id]]
            active_link_id += 1
        
        return net_unit_flux
        
    def get_node_x( self, id ):
        """
        Returns the x coordinate of node "id".
        """
        return self._node_x[id]
        
    def get_node_y( self, id ):
        """
        Returns the y coordinate of node "id".
        """
        return self._node_y[id]

    #Decorator alternatives to above node getters added DEJH late Jul '13.
    #Should allow either .get_property(ID) or .property[ID]
    @property
    def node_x(self):
        return self._node_x
    
    @property
    def node_y(self):
        return self._node_y
        
    def get_cell_x_coords( self ):
        """
        Returns vector of node x coordinates (same as get_node_x_coords).
        
        .. todo: 
            Could be useful to return a Numpy array of x-coords of the cell's
            corners.
        """
        warnings.warn('Use get_node_x_coords instead', DeprecationWarning)
        return self._node_x

    def get_cell_y_coords( self ):
        """
        Returns vector of node y coordinates (same as get_node_y_coords).
        """
        warnings.warn('Use get_node_y_coords instead', DeprecationWarning)
        return self._node_y

    def get_node_x_coords( self ):
        """
        Returns vector of node x coordinates.
        """
        return self._node_x           

    def get_node_y_coords( self ):
        """
        Returns vector of node y coordinates.
        """
        return self._node_y           

    def get_node_coords(self, axis=0):
        """
        Return node coordinates from a given *axis* (defaulting to 0). Axis
        numbering is the same as that for numpy arrays. That is, the zeroth
        axis is along the rows, and the first along the columns.
        """
        assert(axis in (0, 1))

        if axis == 0:
            return self.get_node_y_coords()
        else:
            return self.get_node_x_coords()

    def get_coordinate_units(self, axis=0):
        """
        .. todo:
            GT: coordinate units should be model/component dependent.
        """
        assert(axis in (0, 1))

        if axis == 0:
            return 'degrees_north'
        else:
            return 'degrees_east'

    def get_coordinate_name(self, axis=0):
        """
        .. todo:
            GT: coordinate units should be model/component dependent.
        """
        assert(axis in (0, 1))

        if axis == 0:
            return 'latitude'
        else:
            return 'longitude'

    def get_cell_areas(self, *args):
        """get_cell_areas([ids])

        This function returns the area of an interior cell, or an array of
        all interior cell areas if *ids* is not given. *ids* can be either
        a scalar index, or an array of cell indices.

        .. note::
            It is up to a specific grid class, which has inherited from
            ModelGrid, to construct its own cell_areas array.
        """
        assert(len(args) <= 1)

        if len(args) == 0:
            return self.cell_areas
        else:
            return self.cell_areas[args[0]]
    
    @property
    def cell_areas(self):
        """
        Returns an array of grid-cell areas.

        .. note::
            Sometimes it may make sense for a grid to not always calculate
            its cell areas but, instead, only calculate them once they are
            required. In such cases, the grid class must implement a
            _setup_cell_areas_array method, which will be called the first
            time cell areas are requested.
        """
        try:
            return self.active_cell_areas
        except AttributeError:
            return self._setup_cell_areas_array()

    def get_active_cell_node_ids( self ):
        """
        Returns an integer vector of the node IDs of all active cells.
        """
        return self.activecell_node
        
    def get_active_link_connecting_node_pair(self, node1, node2):
        """
        Returns the ID number of the active link that connects the given pair of
        nodes, or None if not found.
        
        Example:
            
            >>> import landlab as ll
            >>> rmg = ll.RasterModelGrid(4, 5)
            >>> rmg.get_active_link_connecting_node_pair(8, 3)
            2
        """
        active_link = None
        for alink in range(0, self.num_active_links):
            if (self.activelink_fromnode[alink]==node1 \
                and self.activelink_tonode[alink]==node2) \
               or (self.activelink_tonode[alink]==node1
                and self.activelink_fromnode[alink]==node2):
                active_link = alink
                break
        return active_link
        
    def get_minimum_active_link_length(self):
        """
        Returns the horizontal length of the shortest active link in the grid.
        """
        return numpy.amin(self.link_length[self.active_links])

    def assign_upslope_vals_to_active_links( self, u, v=0 ):
        """
        Assigns to each active link the value of u at whichever of its
        neighbors has a higher value of v. If v is omitted, uses u for
        both.
        """
        fv = numpy.zeros( self.num_active_links )
        if len(v) < len(u):
            for i in xrange( 0, self.num_active_links ):
                fv[i] = max( u[self.activelink_fromnode[i]], 
                             u[self.activelink_tonode[i]] )
        else:
            for i in xrange( 0, self.num_active_links ):
                if v[self.activelink_fromnode[i]] > v[self.activelink_tonode[i]]:
                    fv[i] = u[self.activelink_fromnode[i]]
                else:
                    fv[i] = u[self.activelink_tonode[i]]
        return fv
        
    def reset_list_of_active_links(self):
        """
        Creates or resets a list of active links. We do this by sweeping
        through the given lists of from and to nodes, and checking the status
        of these as given in the node_status list. A link is active if both its
        nodes are active interior points, or if one is an active interior and
        the other is an active boundary.
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.reset_list_of_active_links'
            
        fromnode_status = self.node_status[self.link_fromnode]
        tonode_status = self.node_status[self.link_tonode]

        active_links = (((fromnode_status == INTERIOR_NODE) & ~
                         (tonode_status == INACTIVE_BOUNDARY)) |
                        ((tonode_status == INTERIOR_NODE) & ~
                         (fromnode_status == INACTIVE_BOUNDARY)))

        (self.active_links, ) = numpy.where(active_links)

        self.num_active_links = len(self.active_links)
        self.activelink_fromnode = self.link_fromnode[self.active_links]
        self.activelink_tonode = self.link_tonode[self.active_links]
        
        # Set up active inlink and outlink matrices
        self.setup_active_inlink_and_outlink_matrices()

    def deactivate_nodata_nodes(self, node_data, nodata_value):
        """
        Sets self.node_status to INACTIVE_BOUNDARY for all nodes whose value of
        node_data is equal to the nodata_value.
        
        Example:
            
            >>> import landlab as ll
            >>> mg = ll.RasterModelGrid(3, 4, 1.0)
            >>> mg.node_status
            array([1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1], dtype=int8)
            >>> h = numpy.array([-9999,-9999,-9999,-9999,-9999,-9999,12345.,0.,-9999,0.,0.,0.])
            >>> mg.deactivate_nodata_nodes(h, -9999)
            >>> mg.node_status
            array([4, 4, 4, 4, 4, 4, 0, 1, 4, 1, 1, 1], dtype=int8)
        """
        
        # Find locations where value equals the NODATA code and set these nodes
        # as inactive boundaries.
        nodata_locations = numpy.nonzero(node_data==nodata_value)
        self.node_status[nodata_locations] = INACTIVE_BOUNDARY
        
        # Recreate the list of active cell IDs
        node_ids = numpy.array(range(0,self.num_nodes))
        self.activecell_node = node_ids[numpy.where(self.node_status == INTERIOR_NODE)]
        
        # Recreate the list of active links
        self.reset_list_of_active_links()
        
    def active_link_max(self, node_data):
        
        """
        For each active link, finds and returns the maximum value of node_data
        at either of the two ends. Use this, for example, if you want to find
        the maximum value of water depth at linked pairs of nodes (by passing
        in an array of water depth values at nodes).
        
        node_data: a 1D numpy array with length = number of nodes
        returns: a 1D numpy array of maximum values, with length = number of
            active links.
        
        Example:
            
            >>> import landlab as ll
            >>> mg = ll.RasterModelGrid(3, 4, 1.0)
            >>> h = numpy.array([2.,2.,8.,0.,8.,0.,3.,0.,5.,6.,8.,3.])
            >>> mg.active_link_max(h)
            array([ 2.,  8.,  6.,  8.,  8.,  3.,  3.])
        """
        return numpy.maximum(node_data[self.activelink_fromnode],
                             node_data[self.activelink_tonode])
        
    def calculate_link_lengths(self, pts, link_from, link_to):
        """
        Calculates and returns length of links between nodes.
        
        Inputs: pts - Nx2 numpy array containing (x,y) values
                link_from - 1D numpy array containing index numbers of nodes at 
                            starting point ("from") of links
                link_to - 1D numpy array containing index numbers of nodes at 
                          ending point ("to") of links
                          
        Returns: 1D numpy array containing horizontal length of each link
        
        Example:
            
            >>> pts = numpy.array([[0.,0.],[3.,0.],[3.,4.]]) # 3:4:5 triangle
            >>> lfrom = numpy.array([0,1,2])
            >>> lto = numpy.array([1,2,0])
            >>> mg = ModelGrid()
            >>> ll = mg.calculate_link_lengths(pts, lfrom, lto)
            >>> ll
            array([ 3.,  4.,  5.])
        """
        dx = pts[link_to,0]-pts[link_from,0]
        dy = pts[link_to,1]-pts[link_from,1]
        link_length = numpy.sqrt( dx*dx + dy*dy )
        return link_length
        
        
    def calculate_numbers_of_node_neighbors(self):
        """
        Calculates the number of neighboring nodes for each node, and returns
        the result as a 1D numpy array. Used to find the maximum number of
        neighbors, so that inlink and outlink matrices can be dimensioned
        accordingly. Assumes that self.num_nodes, self.link_fromnode, and
        self.link_tonode have already been set up.
        
        Algorithm works by simply looping through all links; for each, the 
        endpoints are neighbors of one another, so we increment the number of
        neighbors for both the endpoint nodes.
        """
        num_nbrs = numpy.zeros(self.num_nodes, dtype=int)
        for link in range(self.num_links):
            num_nbrs[self.link_fromnode[link]] += 1
            num_nbrs[self.link_tonode[link]] += 1
        return num_nbrs


    def setup_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of inlinks and outlinks
        for each node. An inlink of a node is simply a link that has the node as
        its "to" node, and an outlink is a link that has the node as its "from".
        
        We store the inlinks in an NM-row by num_nodes-column matrix called
        node_inlink_matrix. NM is the maximum number of neighbors for any node.
        
        We also keep track of the total number of inlinks and outlinks at each
        node in the num_inlinks and num_outlinks arrays.
        
        The inlink and outlink matrices are useful in numerical calculations.
        Each row of each matrix contains one inlink or outlink per node. So, if
        you have a corresponding "flux" matrix, you can map incoming or
        outgoing fluxes onto the appropriate nodes. More information on this is
        in the various calculate_flux_divergence... functions.
        
        What happens if a given node does not have two inlinks or outlinks? We
        simply put the default value -1 in this case. This allows us to use a 
        cute little trick when computing inflows and outflows. We make our 
        "flux" array one element longer than the number of links, with the last
        element containing the value 0. Thus, any time we add an influx from 
        link number -1, Python takes the value of the last element in the array,
        which is zero. By doing it this way, we maintain the efficiency that 
        comes with the use of numpy. Again, more info can be found in the 
        description of the flux divergence functions.
        
        Example:
            
        """
        
        # Find the maximum number of neighbors for any node
        num_nbrs = self.calculate_numbers_of_node_neighbors()
        self.max_num_nbrs = numpy.amax(num_nbrs)

        # Create active in-link and out-link matrices.
        self.node_inlink_matrix = - numpy.ones((self.max_num_nbrs, self.num_nodes), dtype=numpy.int)
        self.node_outlink_matrix = - numpy.ones((self.max_num_nbrs, self.num_nodes), dtype=numpy.int)

        # Set up the inlink arrays
        tonodes = self.link_tonode
        self.node_numinlink = numpy.bincount(tonodes,
                                                 minlength=self.num_nodes)

        counts = count_repeated_values(self.link_tonode)
        for (count, (tonodes, link_ids)) in enumerate(counts):
            self.node_inlink_matrix[count][tonodes] = link_ids

        # Set up the outlink arrays
        fromnodes = self.link_fromnode
        self.node_numoutlink = numpy.bincount(fromnodes,
                                              minlength=self.num_nodes)
        counts = count_repeated_values(self.link_fromnode)
        for (count, (fromnodes, link_ids)) in enumerate(counts):
            self.node_outlink_matrix[count][fromnodes] = link_ids
                
        
    def setup_active_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of active inlinks and 
        active outlinks for each node. These data structures are equivalent to
        the "regular" inlink and outlink matrices, except that it uses the IDs
        of active links (only).
        """
        # Create active in-link and out-link matrices.
        self.node_active_inlink_matrix = - numpy.ones((self.max_num_nbrs, self.num_nodes),
                                                       dtype=numpy.int)
        self.node_active_outlink_matrix = - numpy.ones((self.max_num_nbrs, self.num_nodes),
                                                        dtype=numpy.int)
        # Set up the inlink arrays
        tonodes = self.activelink_tonode
        self.node_numactiveinlink = numpy.bincount(tonodes,
                                                   minlength=self.num_nodes)

        counts = count_repeated_values(self.activelink_tonode)
        for (count, (tonodes, active_link_ids)) in enumerate(counts):
            self.node_active_inlink_matrix[count][tonodes] = active_link_ids

        # Set up the outlink arrays
        fromnodes = self.activelink_fromnode
        self.node_numactiveoutlink = numpy.bincount(fromnodes,
                                                    minlength=self.num_nodes)
        counts = count_repeated_values(self.activelink_fromnode)
        for (count, (fromnodes, active_link_ids)) in enumerate(counts):
            self.node_active_outlink_matrix[count][fromnodes] = active_link_ids

    
    def display_grid(self, draw_voronoi=False):
        """
        Displays the grid (mainly for purposes of debugging/testing and
        visual examples).
        """
        import matplotlib.pyplot as plt
        
        # Plot nodes, colored by boundary vs interior
        plt.plot(self._node_x[self.interior_nodes], 
                 self._node_y[self.interior_nodes], 'go')
        plt.plot(self._node_x[self.boundary_nodes], 
                 self._node_y[self.boundary_nodes], 'ro')
                 
        # Draw links
        for i in range(self.num_links):
            plt.plot([self._node_x[self.link_fromnode[i]],
                     self._node_x[self.link_tonode[i]]],
                     [self._node_y[self.link_fromnode[i]],
                     self._node_y[self.link_tonode[i]]], 'k-')
                     
        # Draw active links
        for link in self.active_links:
            plt.plot([self._node_x[self.link_fromnode[link]],
                     self._node_x[self.link_tonode[link]]],
                     [self._node_y[self.link_fromnode[link]],
                     self._node_y[self.link_tonode[link]]], 'g-')
                     
        # If caller asked for a voronoi diagram, draw that too
        if draw_voronoi!=None:
            from scipy.spatial import Voronoi, voronoi_plot_2d
            pts = numpy.zeros((self.num_nodes, 2))
            pts[:,0] = self._node_x
            pts[:,1] = self._node_y
            vor = Voronoi(pts)
            voronoi_plot_2d(vor)
                     
        plt.show()
        
        
    def is_boundary(self, ids, boundary_flag=None):
        """
        Check if nodes at given *ids* are boundary nodes. Use the
        *boundary_flag* to specify a particular boundary type status flag.
        """
        if boundary_flag is None:
            return ~ (self.node_status[ids] == INTERIOR_NODE)
        else:
            return self.node_status[ids] == boundary_flag
    
    def assign_boundary_nodes_to_grid_sides(self):
        """
        For each boundary node, determines whether it belongs to the left, 
        right, top or bottom of the grid, based on its distance from the grid's
        centerpoint (mean (x,y) position). Returns lists of nodes on each of 
        the four grid sides. Assumes self.node_status, self.num_nodes, 
        self.boundary_nodes, self._node_x, and self._node_y have been initialized.
        
        Example:

            >>> import landlab as ll
            >>> m = ll.HexModelGrid(5, 3, 1.0)
            >>> [l,r,t,b] = m.assign_boundary_nodes_to_grid_sides()
            >>> l
            array([ 7, 12,  3], dtype=int32)
            >>> r
            array([11, 15,  6], dtype=int32)
            >>> t
            array([16, 18, 17], dtype=int32)
            >>> b
            array([0, 2, 1], dtype=int32)
        """
        # Calculate x and y distance from centerpoint
        dx = self._node_x[self.boundary_nodes] - numpy.mean(self._node_x)
        dy = self._node_y[self.boundary_nodes] - numpy.mean(self._node_y)

        return _sort_points_into_quadrants(dx, dy, self.boundary_nodes)
        
    def set_inactive_boundaries(self, bottom_is_inactive, right_is_inactive, 
                                top_is_inactive, left_is_inactive):
        """
        Handles boundary conditions by setting each of the four sides of the 
        rectangular grid to either 'inactive' or 'active (fixed value)' status.
        Arguments are booleans indicating whether the bottom, right, top, and
        left are inactive (True) or not (False).
        
        For an inactive boundary:
            - the nodes are flagged INACTIVE_BOUNDARY
            - the links between them and the adjacent interior nodes are
              inactive (so they appear on link-based lists, but not
              active_link-based lists)
              
        This means that if you call the calculate_gradients_at_active_links
        method, the inactive boundaries will be ignored: there can be no
        gradients or fluxes calculated, because the links that connect to that
        edge of the grid are not included in the calculation. So, setting a
        grid edge to INACTIVE_BOUNDARY is a convenient way to impose a no-flux
        boundary condition. Note, however, that this applies to the grid as a
        whole, rather than a particular variable that you might use in your
        application. In other words, if you want a no-flux boundary in one
        variable but a different boundary condition for another, then use 
        another method.
        
        The following example sets the top and left boundaries as inactive in a
        four-row by five-column grid that initially has all boundaries active
        and all boundary nodes coded as FIXED_VALUE_BOUNDARY (=1):
        
        >>> import landlab as ll
        >>> rmg = ll.HexModelGrid(5, 3, 1.0) # rows, columns, spacing
        >>> rmg.num_active_links
        30
        >>> rmg.node_status
        array([1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1], dtype=int8)
        >>> rmg.set_inactive_boundaries(False, False, True, True)
        >>> rmg.num_active_links
        21
        >>> rmg.node_status
        array([1, 1, 1, 4, 0, 0, 1, 4, 0, 0, 0, 1, 4, 0, 0, 1, 4, 4, 4], dtype=int8)
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.set_inactive_boundaries'
            
        [left_edge, right_edge, top_edge, bottom_edge] = \
                self.assign_boundary_nodes_to_grid_sides()
            
        if bottom_is_inactive:
            self.node_status[bottom_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[bottom_edge] = FIXED_VALUE_BOUNDARY

        if right_is_inactive:
            self.node_status[right_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[right_edge] = FIXED_VALUE_BOUNDARY
            
        if top_is_inactive:
            self.node_status[top_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[top_edge] = FIXED_VALUE_BOUNDARY

        if left_is_inactive:
            self.node_status[left_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[left_edge] = FIXED_VALUE_BOUNDARY
        
        self.reset_list_of_active_links()
                

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/env python
"""
Unit tests for landlab.model_grid
"""

import unittest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid, BAD_INDEX_VALUE, FIXED_GRADIENT_BOUNDARY


_SPACING = 1.
(_NUM_ROWS, _NUM_COLS) = (4, 5)

class TestRasterModelGrid(unittest.TestCase):
    def setUp(self):
        """
        These tests use a grid that 4x5 nodes::

         15------16------17------18------19
          |       |       |       |       |
          |       |       |       |       |
          |       |       |       |       |
         10------11------12------13------14
          |       |       |       |       |
          |       |       |       |       |   
          |       |       |       |       |
          5-------6-------7-------8-------9
          |       |       |       |       |
          |       |       |       |       |
          |       |       |       |       |
          0-------1-------2-------3-------4
        """
        self.grid = RasterModelGrid(num_rows=_NUM_ROWS, num_cols=_NUM_COLS,
                                    dx=_SPACING)

    def test_create_grid(self):
        """
        Create a raster grid.
        """
        grid = RasterModelGrid(num_rows=4, num_cols=5)

    def test_grid_dimensions(self):
        """
        Check the width of the grid. The x-dimension is the columns, and the
        y-dimension is the rows.
        """
        SPACING = 2.
        (NUM_ROWS, NUM_COLS) = (4, 5)

        grid = RasterModelGrid(num_rows=NUM_ROWS, num_cols=NUM_COLS,
                               dx=SPACING)

        self.assertEqual(grid.get_grid_ydimension(), NUM_ROWS * SPACING)
        self.assertEqual(grid.get_grid_xdimension(), NUM_COLS * SPACING)

    def test_grid_dimensions_default_spacing(self):
        """
        Use the default spacing of 1.
        """
        (NUM_ROWS, NUM_COLS) = (4, 5)

        grid = RasterModelGrid(num_rows=NUM_ROWS, num_cols=NUM_COLS)

        self.assertEqual(grid.get_grid_ydimension(), NUM_ROWS * 1.)
        self.assertEqual(grid.get_grid_xdimension(), NUM_COLS * 1.)

    def test_nodes_around_point(self):
        surrounding_ids = self.grid.get_nodes_around_point(2.1, 1.1)
        self.assertListEqual(
            list(surrounding_ids), [7, 12, 13, 8],
            #list(surrounding_ids), [7, 8, 12, 13],
            'incorrect surrounding ids in test_nodes_around_point')

        surrounding_ids = self.grid.get_nodes_around_point(2.1, .9)
        self.assertListEqual(
            list(surrounding_ids), [2, 7, 8, 3],
            #list(surrounding_ids), [2, 3, 7, 8],
            'incorrect surrounding ids in test_nodes_around_point')

    def test_neighbor_list(self):
        neighbors = self.grid.get_neighbor_list(6)
        assert_array_equal(neighbors, np.array([7, 11, 5, 1]))

        neighbors = self.grid.get_neighbor_list(-1)
        assert_array_equal(neighbors, np.array([-1, -1, -1, -1]))

        neighbors = self.grid.get_neighbor_list([6, -1])
        assert_array_equal(neighbors,
                           np.array([[7, 11, 5, 1], [-1, -1, -1, -1]]))

        neighbors = self.grid.get_neighbor_list()
        assert_array_equal(
            neighbors,
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [ 7, 11,  5,  1], [ 8, 12,  6,  2], [ 9, 13,  7,  3], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [12, 16, 10,  6], [13, 17, 11,  7], [14, 18, 12,  8], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1]]))

    def test_neighbor_list_boundary(self):
        """
        All of the neighbor IDs for a boundary cell are -1.
        """
        neighbors = self.grid.get_neighbor_list(0)

        self.assertEqual(list(neighbors), [-1, -1, -1, -1])

    def test_cell_x(self):
        expected_x = [0., 1., 2., 3., 4.,
                      0., 1., 2., 3., 4.,
                      0., 1., 2., 3., 4.,
                      0., 1., 2., 3., 4.]

        for (cell_id, expected) in zip(xrange(12), expected_x):
            cell_x = self.grid.get_node_x(cell_id)
            self.assertEqual(cell_x, expected)

            cell_x = self.grid.node_x[cell_id]
            self.assertEqual(cell_x, expected)

    def test_cell_y(self):
        expected_y = [0., 0., 0., 0., 0.,
                      1., 1., 1., 1., 1.,
                      2., 2., 2., 2., 2.,
                      3., 3., 3., 3., 3.]

        for (cell_id, expected) in zip(xrange(12), expected_y):
            cell_y = self.grid.get_node_y(cell_id)
            self.assertEqual(cell_y, expected)

            cell_y = self.grid.node_y[cell_id]
            self.assertEqual(cell_y, expected)

    def test_node_x_coordinates(self):
        x_coords = self.grid.get_node_x_coords()

        self.assertEqual(list(x_coords), [0., 1., 2., 3., 4.,
                                          0., 1., 2., 3., 4.,
                                          0., 1., 2., 3., 4.,
                                          0., 1., 2., 3., 4.])

    def test_cell_y_coordinates(self):
        y_coords = self.grid.get_node_y_coords()

        self.assertEqual(list(y_coords), [0., 0., 0., 0., 0.,
                                          1., 1., 1., 1., 1.,
                                          2., 2., 2., 2., 2.,
                                          3., 3., 3., 3., 3.])

    def test_diagonal_list(self):
        diagonals = self.grid.get_diagonal_list(6)
        assert_array_equal(diagonals, np.array([12, 10, 0, 2]))

        diagonals = self.grid.get_diagonal_list(-1)
        assert_array_equal(diagonals, np.array([-1, -1, -1, -1]))

        diagonals = self.grid.get_diagonal_list([6, -1])
        assert_array_equal(diagonals,
                           np.array([[12, 10, 0, 2], [-1, -1, -1, -1]]))

        diagonals = self.grid.get_diagonal_list()
        assert_array_equal(
            diagonals,
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [12, 10,  0,  2], [13, 11,  1,  3], [14, 12,  2,  4], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [17, 15,  5,  7], [18, 16,  6,  8], [19, 17,  7,  9], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1]]))

    def test_diagonal_list_boundary(self):
        diagonals = self.grid.get_diagonal_list(0)
        self.assertEqual(list(diagonals), [-1, -1, -1, -1])

    def test_is_interior(self):
        for cell_id in [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19]:
            self.assertFalse(self.grid.is_interior(cell_id))

        for cell_id in [6, 7, 8, 11, 12, 13]:
            self.assertTrue(self.grid.is_interior(cell_id))

    def test_get_interior_cells(self):
        interiors = self.grid.get_active_cell_node_ids()
        self.assertEqual(list(interiors), [6, 7, 8, 11, 12, 13])

    def test_active_links(self):
        self.assertTrue(isinstance(self.grid.active_links, np.ndarray))
        self.assertEqual(self.grid.num_active_links, 17)

        assert_array_equal(self.grid.active_links,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     19, 20, 21, 22, 23, 24, 25, 26]))

    def test_active_link_fromnode(self):
        assert_array_equal(self.grid.activelink_fromnode,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     5, 6, 7, 8, 10, 11, 12, 13]))

    def test_active_link_tonode(self):
        assert_array_equal(self.grid.activelink_tonode,
                           np.array([6, 7, 8, 11, 12, 13, 16, 17, 18,
                                     6, 7, 8, 9, 11, 12, 13, 14]))

    def test_active_link_num_inlink(self):
        assert_array_equal(self.grid.node_numactiveinlink,
                           np.array([0, 0, 0, 0, 0,
                                     0, 2, 2, 2, 1,
                                     0, 2, 2, 2, 1,
                                     0, 1, 1, 1, 0]))

    def test_active_link_num_outlink(self):
        assert_array_equal(self.grid.node_numactiveoutlink,
                           np.array([0, 1, 1, 1, 0,
                                     1, 2, 2, 2, 0,
                                     1, 2, 2, 2, 0,
                                     0, 0, 0, 0, 0]))

    def test_active_inlink_matrix(self):
        self.assertTupleEqual(self.grid.node_active_inlink_matrix.shape,
                              (2, 20))

        assert_array_equal(self.grid.node_active_inlink_matrix[0],
                           np.array([-1, -1, -1, -1, -1,
                                     -1,  0,  1,  2, 12,
                                     -1,  3,  4,  5, 16,
                                     -1,  6,  7,  8, -1]))
        assert_array_equal(self.grid.node_active_inlink_matrix[1],
                           np.array([-1, -1, -1, -1, -1,
                                     -1,  9, 10, 11, -1,
                                     -1, 13, 14, 15, -1,
                                     -1, -1, -1, -1, -1]))

    def test_active_outlink_matrix(self):
        self.assertTupleEqual(self.grid.node_active_outlink_matrix.shape,
                              (2, 20))

        assert_array_equal(self.grid.node_active_outlink_matrix[0],
                           np.array([-1,  0,  1,  2, -1,
                                      9,  3,  4,  5, -1,
                                     13,  6,  7,  8, -1,
                                     -1, -1, -1, -1, -1]))
        assert_array_equal(self.grid.node_active_outlink_matrix[1],
                           np.array([-1, -1, -1, -1, -1,
                                     -1, 10, 11, 12, -1,
                                     -1, 14, 15, 16, -1,
                                     -1, -1, -1, -1, -1]))

    def test_link_fromnode(self):
        assert_array_equal(self.grid.link_fromnode,
                           np.array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
                                     10, 11, 12, 13, 14,  0,  1,  2,  3,  5,
                                      6,  7,  8, 10, 11, 12, 13, 15, 16, 17,
                                     18]))

    def test_link_tonode(self):
        assert_array_equal(self.grid.link_tonode,
                           np.array([ 5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
                                     15, 16, 17, 18, 19,  1,  2,  3,  4,  6,
                                      7,  8,  9, 11, 12, 13, 14, 16, 17, 18,
                                     19]))

    def test_link_num_inlink(self):
        assert_array_equal(self.grid.node_numinlink,
                           np.array([0, 1, 1, 1, 1,
                                     1, 2, 2, 2, 2,
                                     1, 2, 2, 2, 2,
                                     1, 2, 2, 2, 2]))

    def test_link_num_outlink(self):
        assert_array_equal(self.grid.node_numoutlink,
                           np.array([2, 2, 2, 2, 1,
                                     2, 2, 2, 2, 1,
                                     2, 2, 2, 2, 1,
                                     1, 1, 1, 1, 0]))

    def test_node_inlink_matrix(self):
        self.assertTupleEqual(self.grid.node_inlink_matrix.shape,
                              (2, 20))
        assert_array_equal(self.grid.node_inlink_matrix[0],
                           np.array([-1, 15, 16, 17, 18,
                                      0,  1,  2,  3,  4,
                                      5,  6,  7,  8,  9,
                                     10, 11, 12, 13, 14]))
        assert_array_equal(self.grid.node_inlink_matrix[1],
                           np.array([-1, -1, -1, -1, -1,
                                     -1, 19, 20, 21, 22,
                                     -1, 23, 24, 25, 26,
                                     -1, 27, 28, 29, 30]))

    def test_node_outlink_matrix(self):
        self.assertTupleEqual(self.grid.node_outlink_matrix.shape,
                              (2, 20))
        assert_array_equal(self.grid.node_outlink_matrix[0],
                           np.array([ 0,  1,  2,  3,  4,
                                     5,  6,  7,  8,  9,
                                     10, 11, 12, 13, 14,
                                     27, 28, 29, 30, -1]))
        assert_array_equal(self.grid.node_outlink_matrix[1],
                           np.array([15, 16, 17, 18, -1,
                                     19, 20, 21, 22, -1,
                                     23, 24, 25, 26, -1,
                                     -1, -1, -1, -1, -1]))

    def test_link_face(self):
        BAD = BAD_INDEX_VALUE

        assert_array_equal(self.grid.link_face,
                           np.array([BAD, 0, 1, 2, BAD,
                                     BAD, 3, 4, 5, BAD,
                                     BAD, 6, 7, 8, BAD,
                                     BAD, BAD, BAD, BAD,
                                     9, 10, 11, 12,
                                     13, 14, 15, 16,
                                     BAD, BAD, BAD, BAD]))

    def test_grid_coords_to_node_id(self):
        self.assertEqual(self.grid.grid_coords_to_node_id(3, 4), 19)
        assert_array_equal(self.grid.grid_coords_to_node_id((3, 2), (4, 1)),
                           np.array([19, 11]))

        with self.assertRaises(ValueError):
            self.grid.grid_coords_to_node_id(5, 0)

    def test_create_diagonal_list(self):
        self.grid.create_diagonal_list()

        assert_array_equal(
            self.grid.get_diagonal_list(),
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [12, 10,  0,  2], [13, 11,  1,  3],
                      [14, 12,  2,  4], [-1, -1, -1, -1], [-1, -1, -1, -1], [17, 15,  5,  7],
                      [18, 16,  6,  8], [19, 17,  7,  9], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1]]))


class TestZerosArray(unittest.TestCase):
    def test_default(self):
        mg = RasterModelGrid(4, 5)
        v = mg.zeros()
        assert_array_equal(v, np.zeros(20., dtype=np.float))

    def test_int_array(self):
        mg = RasterModelGrid(4, 5)
        v = mg.zeros(dtype=np.int)
        assert_array_equal(v, np.zeros(20., dtype=np.int))

    def test_at_cells(self):
        mg = RasterModelGrid(4, 5)
        v = mg.zeros(centering='cell')
        assert_array_equal(v, np.zeros(6, dtype=np.float))

    def test_at_links(self):
        mg = RasterModelGrid(4, 5)
        v = mg.zeros(centering='link')
        assert_array_equal(v, np.zeros(17, dtype=np.float))

    def test_at_faces(self):
        mg = RasterModelGrid(4, 5)
        v = mg.zeros(centering='face')
        assert_array_equal(v, np.zeros(17, dtype=np.float))


class TestRasterModelGridZeroArrays(unittest.TestCase):
    def test_default(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(), np.zeros((20, ), dtype=np.float))

    def test_at_nodes(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(centering='node'),
                           np.zeros((20, ), dtype=np.float))

    def test_at_cells(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(centering='cell'),
                           np.zeros((6, ), dtype=np.float))

    def test_at_links(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(centering='link'),
                           np.zeros((17, ), dtype=np.float))

    def test_at_faces(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(centering='face'),
                           np.zeros((17, ), dtype=np.float))

    def test_bad_centering(self):
        mg = RasterModelGrid(4, 5)
        with self.assertRaises(TypeError):
            mg.zeros(centering='bad_centering_string')

    def test_int(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(dtype=np.int),
                           np.zeros((20, ), dtype=np.int))

    def test_int8(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(mg.zeros(dtype=np.int8),
                           np.zeros((20, ), dtype=np.int8))

class TestRasterModelGridIsBoundary(unittest.TestCase):
    def test_id_as_int(self):
        mg = RasterModelGrid(4, 5)
        self.assertTrue(mg.is_boundary(0))

    def test_id_as_small_list(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.is_boundary([0]),
            np.array([True]))

    def test_id_as_array(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.is_boundary(np.arange(20)),
            np.array([True,  True,  True,  True,  True,
                      True, False, False, False,  True,  
                      True, False, False, False,  True,  
                      True,  True,  True,  True,  True], dtype=bool))

    def test_id_as_list(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.is_boundary([8, 9]),
            np.array([False, True]))

    def test_boundary_flag(self):
        mg = RasterModelGrid(4, 5)
        mg.node_status[0] = FIXED_GRADIENT_BOUNDARY
        assert_array_equal(
            mg.is_boundary(np.arange(20)),
            np.array([True,  True,  True,  True,  True,
                      True, False, False, False,  True,  
                      True, False, False, False,  True,  
                      True,  True,  True,  True,  True], dtype=bool))

        assert_array_equal(
            mg.is_boundary(np.arange(20), boundary_flag=FIXED_GRADIENT_BOUNDARY),
            np.array([ True, False, False, False, False,
                      False, False, False, False, False,
                      False, False, False, False, False,
                      False, False, False, False, False], dtype=bool))


class TestRasterModelGridHasBoundaryNeighbor(unittest.TestCase):
    def test_boundary_node(self):
        mg = RasterModelGrid(5, 6)
        self.assertTrue(mg.has_boundary_neighbor(0))
        self.assertFalse(mg.has_boundary_neighbor(14))

    def test_last_index(self):
        mg = RasterModelGrid(4, 5)
        self.assertTrue(mg.has_boundary_neighbor(-1))

    def test_id_as_list(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.has_boundary_neighbor([-1, 0]),
            np.array([True, True])
        )

    def test_id_as_array(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.has_boundary_neighbor(np.arange(20)),
            np.array(
                [True, True, True, True, True,
                 True, True, True, True, True,
                 True, True, True, True, True,
                 True, True, True, True, True]))

        mg = RasterModelGrid(5, 5)
        assert_array_equal(
            mg.has_boundary_neighbor(np.arange(25)),
            np.array(
                [True, True,  True, True, True,
                 True, True,  True, True, True,
                 True, True, False, True, True,
                 True, True,  True, True, True,
                 True, True,  True, True, True]))


class TestRasterModelGridConnectingFaces(unittest.TestCase):
    def test_horizontally_adjacent_cells(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.get_face_connecting_cell_pair(0, 1),
            np.array([10]))

    def test_vertically_adjacent_cells(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.get_face_connecting_cell_pair(0, 3),
            np.array([3]))

    def test_diagonally_adjacent_cells(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.get_face_connecting_cell_pair(1, 5),
            np.array([]))

    def test_non_adjacent_cells(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.get_face_connecting_cell_pair(0, 2),
            np.array([]))


class TestRasterModelGridCellFaces(unittest.TestCase):
    def test_id_as_int(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.cell_faces(0),
            np.array([0, 9, 3, 10]))

    def test_id_as_array(self):
        mg = RasterModelGrid(4, 5)
        assert_array_equal(
            mg.cell_faces(np.array([0, 1])),
            np.array([[0, 9, 3, 10],
                      [1, 10, 4, 11]]))


class TestRasterModelGridNodesAroundPoint(unittest.TestCase):
    def test_lower_left_cell(self):
        mg = RasterModelGrid(3, 3)
        assert_array_equal(
            mg.get_nodes_around_point(.1, .1),
            np.array([0, 3, 4, 1]))
        assert_array_equal(
            mg.get_nodes_around_point(.1, .9),
            np.array([0, 3, 4, 1]))
        assert_array_equal(
            mg.get_nodes_around_point(.9, .9),
            np.array([0, 3, 4, 1]))
        assert_array_equal(
            mg.get_nodes_around_point(.9, .1),
            np.array([0, 3, 4, 1]))


class TestRasterModelGridCellAreas(unittest.TestCase):
    def test_unit_grid_all_cells(self):
        mg = RasterModelGrid(4, 4)
        assert_array_equal(mg.get_cell_areas(), np.ones(4.))

    def test_unit_grid_one_cell(self):
        mg = RasterModelGrid(4, 4)
        self.assertEqual(mg.get_cell_areas(0), 1.)

    def test_unit_grid_last_cell(self):
        mg = RasterModelGrid(4, 4)
        self.assertEqual(mg.get_cell_areas(-1), 1.)

    def test_out_of_range(self):
        mg = RasterModelGrid(4, 4)
        with self.assertRaises(IndexError):
            mg.get_cell_areas(5)

    def test_all_cells_with_spacing(self):
        mg = RasterModelGrid(4, 4, 10.)
        assert_array_equal(mg.get_cell_areas(), 100. * np.ones(4.))


if __name__ == '__main__':
    unittest.main()



########## soil_moisture.py ###############
##
##  This component calculates and updates soil moisture after each storm.
##  Multiple storms can be considered which is fed as an input.
##  This code implements Laio's(2001) solution for soil moisture calculation.
##  Input file is named soilmoisture_input.txt and is temporarily placed under landlab.components.
##
##  Storms are considered to be instantaneous events.
##  Storm duration, depth and interstorm duration are obtained from RainfallDriver.py
##  Storm depth is obtained in mm and storm duration and interstorm duration are obtained in hours
##
##  After every storm, Laio's solution for soil moisture drying is implemented. 
##
##
## Written by Erkan Istanbulluoglu & Sai Nudurupati, 2013.
###########################################

import os
from numpy import *
from math import *
import landlab
from landlab.components.RainfallDriver import PrecipitationDistribution
from landlab.model_parameter_dictionary import ModelParameterDictionary
from matplotlib.pyplot import *


_DEFAULT_INPUT_FILE = os.path.join(os.path.dirname(__file__),
                                  'soilmoisture_input.txt')

class SoilMoisture():

    def __init__( self ):
        """ vegcover - vegetation cover fraction """
        self._vegcover = 0.0
        """ interception_cap - Maximum interception capacity """
        self._interception_cap = 0.0
        """ zr - root depth """
        self._zr = 0.0
        """ runon - runon from upstream """
        self._runon = 0.0
        """ PET - potential evapotranspiration """
        self._PET = 0.0
        """ fbare - scaling coefficient for bare soil Potential Evapotranspiration """
        self._fbare = 0.0

        """ soil_type - Soil type to be selected """
        self._soil_type = 1
        """ Ib - Bare soil infiltration capacity [mm/h] """
        self._soil_Ib = 0.0
        """ Iv - Vegetated surface infiltration capacity [mm/h] """
        self._soil_Iv = 0.0
        """ Ew - Residual Evaporation after wilting """
        self._soil_Ew = 0.0
        """ pc - porosity """
        self._soil_pc = 0.0
        """ fc - field capacity """
        self._soil_fc = 0.0
        """ sc - stomata closure moisture fraction """
        self._soil_sc = 0.0
        """ wp - wilting point """ 
        self._soil_wp = 0.0
        """ hgw - Hygroscopic water """
        self._soil_hgw = 0.0
        """ beta - Deep percolation rate constant """
        self._soil_beta = 0.0
        
        """ P - Precipitation """
        P = 0.0
        """ ETA - Evapotranspiration """
        ETA = 0.0
        """ sini - soil moisture with storm at the end of time intervals """
        sini = 0.0
        """ s - final soil moisture """
        s = 0.0
        """ Dd - Drainage from rootzone """
        Dd = 0.0
        """ Tb - Time between storms """
        Tb = 0.0
        """ Tr - Storm duration """
        Tr = 0.0
        """ Runoff - runoff after storm """
        Runoff = 0.0

        """ Iterations - Number of iterations/storms """
        self._iterate_storm = 0
        """ so - Soil moisture of the ground before storm """
        so = 0.0

    def initialize( self ):

        MPD = ModelParameterDictionary()
        MPD.read_from_file(_DEFAULT_INPUT_FILE)

        self._iterate_storm = MPD.read_int( 'strms' )
        print '\nNumber of storms: ', self._iterate_storm
        self._vegcover = MPD.read_float( 'VegCover' )
        print '\n Veg cover: ', self._vegcover
        self._interception_cap = MPD.read_float( 'InterceptionCap' )
        print '\n Interception cap', self._interception_cap
        self._zr = MPD.read_float( 'ZR' )
        print '\n Root depth', self._zr
        self._runon = MPD.read_float( 'Runon' )
        print '\n Initial runoff:', self._runon
        self._PET = MPD.read_float( 'PET' )
        print '\n Fixed PET: ', self._PET
        self._fbare = MPD.read_float( 'F_Bare' )
        print '\n F_bare: ', self._fbare
        self._so = MPD.read_float( 'so' )        
        self._si = self._so
        print '\n Initial soil moisture: ', self._so
        
        self._soil_type = MPD.read_float( 'SoilType' )
        self._soil_Ib = MPD.read_float( 'Ib' )
        print '\n Ib: ', self._soil_Ib
        self._soil_Iv = MPD.read_float( 'Iv' )
        print '\n Iv: ', self._soil_Iv
        self._soil_Ew = MPD.read_float( 'Ew' )
        print '\n Ew: ', self._soil_Ew
        self._soil_pc = MPD.read_float( 'pc' )
        print '\n pc: ', self._soil_pc
        self._soil_fc = MPD.read_float( 'fc' )
        print '\n fc: ', self._soil_fc
        self._soil_sc = MPD.read_float( 'sc' )
        print '\n sc: ', self._soil_sc
        self._soil_wp = MPD.read_float( 'wp' )
        print '\n wp: ', self._soil_wp
        self._soil_hgw = MPD.read_float( 'hgw' )
        print '\n hgw: ', self._soil_hgw
        self._soil_beta = MPD.read_float( 'Beta' )
        print '\n Beta: ', self._soil_beta
        
        self._length = self._iterate_storm
        """ Initializing arrays to store values of variables """
        self._Pin = zeros( self._length, dtype = float )
        self._Pa = zeros( self._length, dtype = float )
        self._P = zeros( self._length, dtype = float )
        self._ETA = zeros( self._length, dtype = float )
        self._Sini = zeros( self._length, dtype = float )
        self._S = zeros( self._length, dtype = float )
        self._D = zeros( self._length, dtype = float )
        self._Tb = zeros( self._length, dtype = float )
        self._Tr = zeros( self._length, dtype = float )
        self._Ro = zeros( self._length, dtype = float )
        self._tfc = zeros( self._length, dtype = float )
        self._tsc = zeros( self._length, dtype = float )
        self._twp = zeros( self._length, dtype = float )

        self._Total = zeros( self._length, dtype = float )
        self._percentage = zeros( self._length, dtype = float )

    def update( self ):

        V = self._vegcover
        fb = self._fbare
        ZR = self._zr
        fbare = self._fbare
        PET = self._PET
        pc = self._soil_pc
        fc = self._soil_fc
        sc = self._soil_sc
        wp = self._soil_wp
        hgw = self._soil_hgw
        beta = self._soil_beta
           
        """ Create an instance of Storm Class """
        PD = PrecipitationDistribution()
        """ Initialize Storm - Create first storm """
        PD.initialize()
        
        for i in range( 0, self._iterate_storm ):
            
           if i != 0:
               """ Create a new storm """ 
               PD.update()

           """ Soil moisture dynamics """
           P = PD.storm_depth
           Pin = PD.intensity
           Tb = PD.interstorm_duration
           Tr = PD.storm_duration                      
           
           Inf_cap = self._soil_Ib*(1-V) + self._soil_Iv*V
           Int_cap = min(V*self._interception_cap, P)
           Peff = max(P-Int_cap, 0.0)
           mu = (Int_cap/1000.0)/(pc*ZR*(exp(beta*(1-fc))-1))
           Ep = max((PET*V+fbare*PET*(1-V)) - Int_cap, 0.0)
           nu = ((Ep/24.0)/1000.0)/(pc*ZR)
           nuw = ((Ep*0.1/24)/1000.0)/(pc*ZR)

           sini = self._so + (Peff/(pc*ZR*1000.0))+self._runon

           if sini>1:
               runoff = (sini-1)*pc*ZR*1000
               sini = 1
           else:
               runoff = 0

           if sini>=fc:
                tfc = (1.0/(beta*(mu-nu)))*(beta*(fc-sini)+log((nu-mu+mu*exp(beta*(sini-fc)))/nu))
                tsc = ((fc-sc)/nu)+tfc
                twp = ((sc-wp)/(nu-nuw))*log(nu/nuw)+tsc

                if Tb<tfc:
                    s = abs(sini-(1/beta)*log(((nu-mu+mu*exp(beta*(sini-fc)))*exp(beta*(nu-mu)*Tb)-mu*exp(beta*(sini-fc)))/(nu-mu)))
                    Dd = ((pc*ZR*1000)*(sini-s))-(Tb*(Ep/24))
                    ETA = (Tb*(Ep/24))

                elif Tb>=tfc and Tb<tsc:
                    s = fc-(nu*(Tb-tfc))
                    Dd = ((pc*ZR*1000)*(sini-fc))-((tfc)*(Ep/24))
                    ETA = (Tb*(Ep/24))

                elif Tb>=tsc and Tb<twp:
                    s = wp+(sc-wp)*((nu/(nu-nuw))*exp((-1)*((nu-nuw)/(sc-wp))*(Tb-tsc))-(nuw/(nu-nuw)))
                    Dd = ((pc*ZR*1000)*(sini-fc))-(tfc*Ep/24)
                    ETA = (1000*ZR*pc*(sini-s))-Dd

                else:
                    s = hgw+(wp-hgw)*exp((-1)*(nuw/(wp-hgw))*max(Tb-twp,0))
                    Dd = ((pc*ZR*1000)*(sini-fc))-(tfc*Ep/24)
                    ETA = (1000*ZR*pc*(sini-s))-Dd

           elif sini<fc and sini>=sc:
                tfc = 0
                tsc = (sini-sc)/nu
                twp = ((sc-wp)/(nu-nuw))*log(nu/nuw)+tsc

                if Tb<tsc:
                    s = sini - nu*Tb
                    Dd = 0
                    ETA = 1000*ZR*pc*(sini-s)

                elif Tb>=tsc and Tb<twp:
                    s = wp+(sc-wp)*((nu/(nu-nuw))*exp((-1)*((nu-nuw)/(sc-wp))*(Tb-tsc))-(nuw/(nu-nuw)))
                    Dd = 0
                    ETA = (1000*ZR*pc*(sini-s))

                else:
                    s = hgw+(wp-hgw)*exp((-1)*(nuw/(wp-hgw))*(Tb-twp))
                    Dd = 0
                    ETA = (1000*ZR*pc*(sini-s))
                    
           elif sini<sc and sini>=wp:
                tfc = 0
                tsc = 0
                twp = ((sc-wp)/(nu-nuw))*log(1+(nu-nuw)*(sini-wp)/(nuw*(sc-wp)))

                if Tb<twp:
                    s = wp+((sc-wp)/(nu-nuw))*((exp((-1)*((nu-nuw)/(sc-wp))*Tb))*(nuw+((nu-nuw)/(sc-wp))*(sini-wp))-nuw)
                    Dd = 0
                    ETA = (1000*ZR*pc*(sini-s))

                else:
                    s = hgw+(wp-hgw)*exp((-1)*(nuw/(wp-hgw))*(Tb-twp))
                    Dd = 0
                    ETA = (1000*ZR*pc*(sini-s))

           else:
                tfc = 0
                tsc = 0
                twp = 0

                s = hgw+(sini-hgw)*exp((-1)*(nuw/(wp-hgw))*Tb)
                Dd = 0
                ETA = (1000*ZR*pc*(sini-s))

           self._Pa[i] = P
           self._Pin[i] = Pin
           self._P[i] = Peff
           self._ETA[i] = ETA
           self._Sini[i] = sini
           self._S[i] = s
           self._D[i] = Dd
           self._Tb[i] = Tb
           self._Tr[i] = Tr
           self._Ro[i] = runoff
           self._tfc[i] = tfc
           self._tsc[i] = tsc
           self._twp[i] = twp

           self._so = s

    def plott( self ):

        figure(1)
        self._px = xrange( 0, self._iterate_storm )
        plot( self._px, self._P )
        ylabel( 'Effective Precipitation Depth in mm' )
        title( 'Precipitation Record' )
        xlabel( 'Storms' )

        figure(2)
        plot( self._px, self._S )
        ylabel( 'Soil Moisture - fraction' )
        title( 'Soil Moisture Dynamics' )
        xlabel( 'Storms' )

        figure(3)
        plot( self._px, self._ETA, '+', label = 'Evapotranspiration' )
        hold(True)
        plot( self._px, self._D, '-', label = 'Drainage from Root Zone' )
        hold(True)
        plot( self._px, self._Ro, '*', label = 'Runoff' )
        hold(False)
        xlabel( 'Storms' )
        ylabel( 'Depth in mm' )
        legend()
        title( 'Leakage/Losses' )
                
        show()



        



#! /usr/bin/env python

import unittest
import numpy as np
from numpy.testing import assert_array_equal

import landlab.utils.structured_grid as sgrid
from landlab.utils.structured_grid import BAD_INDEX_VALUE


class TestGetNodeCoords(unittest.TestCase):
    def test_node_x_2d(self):
        (x, _) = sgrid.node_coords((3,2))

        assert_array_equal(x, np.array([0., 1.,
                                        0., 1.,
                                        0., 1.,]))

    def test_node_x_2d_with_spacing(self):
        (x, _) = sgrid.node_coords((3,2), (2., 10.))

        assert_array_equal(x, np.array([0., 10.,
                                        0., 10.,
                                        0., 10.,]))

    def test_node_x_2d_with_origin(self):
        (x, _) = sgrid.node_coords((3,2), (2., 10.), (-1., 1.))

        assert_array_equal(x, np.array([1., 11.,
                                        1., 11.,
                                        1., 11.,]))

    def test_node_y_2d(self):
        (_, y) = sgrid.node_coords((3,2))

        assert_array_equal(y, np.array([0., 0.,
                                        1., 1.,
                                        2., 2.,]))

    def test_node_y_2d_with_spacing(self):
        (_, y) = sgrid.node_coords((3,2), (2., 10.))

        assert_array_equal(y, np.array([0., 0.,
                                        2., 2.,
                                        4., 4.,]))

    def test_node_y_2d_with_origin(self):
        (_, y) = sgrid.node_coords((3,2), (2., 10.), (-1., 1.))

        assert_array_equal(y, np.array([-1., -1.,
                                         1.,  1.,
                                         3.,  3.,]))


class TestGetCellNode(unittest.TestCase):
    def test_2d_shape_2_by_3(self):
        cell_nodes = sgrid.cell_node_index((2, 3))

        assert_array_equal(cell_nodes, np.array([]))

    def test_2d_shape_3_by_3(self):
        cell_nodes = sgrid.cell_node_index((3, 3))

        assert_array_equal(cell_nodes, np.array([4]))

    def test_shape_4_by_5(self):
        cell_nodes = sgrid.cell_node_index((4, 5))

        assert_array_equal(cell_nodes, np.array([ 6,  7,  8,
                                                 11, 12, 13]))


class TestGetNodeLinks(unittest.TestCase):
    def test_2d_3_by_2_from_links(self):
        (from_indices, _) = sgrid.node_link_index((3, 2))

        assert_array_equal(from_indices,
                           np.array([0, 1, 2, 3,
                                     0, 2, 4]))

    def test_2d_3_by_2_to_links(self):
        (_, to_indices) = sgrid.node_link_index((3, 2))

        assert_array_equal(to_indices,
                           np.array([2, 3, 4, 5,
                                     1, 3, 5]))


class TestNodeActiveCell(unittest.TestCase):
    def test_one_active_cell(self):
        active_cells = sgrid.node_active_cell((3, 3))

        assert_array_equal(
            active_cells,
            np.array([BAD_INDEX_VALUE, BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                      BAD_INDEX_VALUE,               0, BAD_INDEX_VALUE,
                      BAD_INDEX_VALUE, BAD_INDEX_VALUE, BAD_INDEX_VALUE])
        )

    def test_no_active_cells(self):
        active_cells = sgrid.node_active_cell((3, 2))

        assert_array_equal(
            active_cells,
            np.array([BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                      BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                      BAD_INDEX_VALUE, BAD_INDEX_VALUE])
        )
                                     
                                     
class TestActiveCells(unittest.TestCase):
    def test_one_active_cell(self):
        active_cells = sgrid.active_cells((3, 3))

        assert_array_equal(active_cells, np.array([0]))

    def test_no_active_cells(self):
        active_cells = sgrid.active_cells((3, 2))

        assert_array_equal(active_cells,
            np.array([]))


class TestCellCount(unittest.TestCase):
    def test_one_cell(self):
        n_cells = sgrid.cell_count((3, 3))
        self.assertEqual(n_cells, 1)

    def test_no_cells(self):
        n_cells = sgrid.cell_count((2, 3))
        self.assertEqual(n_cells, 0)


class TestInteriorCellCount(unittest.TestCase):
    def test_one_cell(self):
        n_cells = sgrid.interior_cell_count((3, 3))
        self.assertEqual(n_cells, 1)

    def test_no_cells(self):
        n_cells = sgrid.interior_cell_count((2, 3))
        self.assertEqual(n_cells, 0)


class TestActiveCellCount(unittest.TestCase):
    def test_one_cell(self):
        n_cells = sgrid.active_cell_count((3, 3))
        self.assertEqual(n_cells, 1)

    def test_no_cells(self):
        n_cells = sgrid.active_cell_count((2, 3))
        self.assertEqual(n_cells, 0)


class TestInteriorNodes(unittest.TestCase):
    def test_4_by_5(self):
        interiors = sgrid.interior_nodes((4, 5))
        assert_array_equal(interiors, np.array([6, 7, 8, 11, 12, 13]))

    def test_no_interiors(self):
        interiors = sgrid.interior_nodes((2, 3))
        assert_array_equal(interiors, np.array([]))


class TestNodeStatus(unittest.TestCase):
    def test_4_by_5(self):
        status = sgrid.node_status((4, 5))
        self.assertEqual(status.dtype, np.int8)
        assert_array_equal(status,
                           np.array([1, 1, 1, 1, 1,
                                     1, 0, 0, 0, 1,
                                     1, 0, 0, 0, 1,
                                     1, 1, 1, 1, 1, ]))

    def test_no_interiors(self):
        status = sgrid.node_status((2, 3))
        self.assertEqual(status.dtype, np.int8)
        assert_array_equal(status,
                           np.array([1, 1, 1,
                                     1, 1, 1,]))


class TestActiveLinks(unittest.TestCase):
    """
    *--27-->*--28-->*--29-->*--30-->*
    ^       ^       ^       ^       ^
    10      11      12      13      14
    |       |       |       |       |
    *--23-->*--24-->*--25-->*--26-->*
    ^       ^       ^       ^       ^
    5       6       7       8       9   
    |       |       |       |       |
    *--19-->*--20-->*--21-->*--22-->*
    ^       ^       ^       ^       ^
    0       1       2       3       4
    |       |       |       |       |
    *--15-->*--16-->*--17-->*--18-->*
    """
    def test_4_by_5(self):
        active_links = sgrid.active_links((4, 5))
        assert_array_equal(active_links,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     19, 20, 21, 22, 23, 24, 25, 26]))
        self.assertEqual(len(active_links), sgrid.active_link_count((4, 5)))

    def test_with_node_status(self):
        status = sgrid.node_status((4, 5))
        status[6] = sgrid.INACTIVE_BOUNDARY
        active_links = sgrid.active_links((4, 5), node_status_array=status)

        assert_array_equal(active_links,
                           np.array([2, 3, 7, 8, 11, 12, 13,
                                     21, 22, 23, 24, 25, 26]))

    def test_with_link_nodes(self):
        link_nodes = sgrid.node_link_index((4, 5))
        active_links = sgrid.active_links((4, 5), link_nodes=link_nodes)

        assert_array_equal(active_links,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     19, 20, 21, 22, 23, 24, 25, 26]))
        self.assertEqual(len(active_links), sgrid.active_link_count((4, 5)))


class TestLinkFaces(unittest.TestCase):
    def test_4_by_5(self):
        link_faces = sgrid.link_faces((4, 5))

        BAD = sgrid.BAD_INDEX_VALUE

        assert_array_equal(link_faces, np.array([BAD, 0, 1, 2, BAD,
                                                 BAD, 3, 4, 5, BAD,
                                                 BAD, 6, 7, 8, BAD,
                                                 BAD, BAD, BAD, BAD,
                                                 9, 10, 11, 12,
                                                 13, 14, 15, 16,
                                                 BAD, BAD, BAD, BAD]))

    def test_with_active_links(self):
        active_links = sgrid.active_links((4, 5))
        active_links = active_links[:-1]
        link_faces = sgrid.link_faces((4, 5), actives=active_links)

        BAD = sgrid.BAD_INDEX_VALUE

        assert_array_equal(link_faces, np.array([BAD, 0, 1, 2, BAD,
                                                 BAD, 3, 4, 5, BAD,
                                                 BAD, 6, 7, 8, BAD,
                                                 BAD, BAD, BAD, BAD,
                                                 9, 10, 11, 12,
                                                 13, 14, 15, BAD,
                                                 BAD, BAD, BAD, BAD]))


class TestReshapeArray(unittest.TestCase):
    def test_default(self):
        x = np.arange(12.)
        y = sgrid.reshape_array((3, 4), x)

        self.assertEqual(y.shape, (3, 4))
        assert_array_equal(x, y.flat)
        self.assertTrue(y.flags['C_CONTIGUOUS'])
        self.assertIs(y.base, x)

    def test_copy(self):
        x = np.arange(12.)
        y = sgrid.reshape_array((3, 4), x, copy=True)

        self.assertEqual(y.shape, (3, 4))
        assert_array_equal(x, y.flat)
        self.assertTrue(y.flags['C_CONTIGUOUS'])
        self.assertIsNone(y.base)

        y[0][0] = 0.
        assert_array_equal(x, np.array([ 0.,  1.,  2., 3.,
                                         4.,  5.,  6., 7.,
                                         8.,  9., 10., 11.]))

    def test_flip(self):
        x = np.arange(12.)
        y = sgrid.reshape_array((3, 4), x, flip_vertically=True)

        self.assertEqual(y.shape, (3, 4))
        assert_array_equal(y, np.array([[ 8.,  9., 10., 11.],
                                        [ 4.,  5.,  6., 7.],
                                        [ 0.,  1.,  2., 3.]]))
        self.assertFalse(y.flags['C_CONTIGUOUS'])
        self.assertIsNotNone(y.base)

        y[0][0] = 0.
        assert_array_equal(x, np.array([ 0.,  1.,  2., 3.,
                                         4.,  5.,  6., 7.,
                                         0.,  9., 10., 11.]))

    def test_flip_copy(self):
        x = np.arange(12.)
        y = sgrid.reshape_array((3, 4), x, flip_vertically=True, copy=True)

        self.assertEqual(y.shape, (3, 4))
        assert_array_equal(y, np.array([[ 8.,  9., 10., 11.],
                                        [ 4.,  5.,  6., 7.],
                                        [ 0.,  1.,  2., 3.]]))
        self.assertTrue(y.flags['C_CONTIGUOUS'])
        self.assertIsNot(y.base, x)


class TestDiagonalArray(unittest.TestCase):
    def test_default(self):
        diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1)
        assert_array_equal(diags,
                           np.array([[ 4, -1, -1, -1],
                                     [ 5,  3, -1, -1],
                                     [-1,  4, -1, -1],
                                     [-1, -1, -1,  1],
                                     [-1, -1,  0,  2],
                                     [-1, -1,  1, -1]]))

        self.assertTrue(diags.base is None)
        self.assertTrue(diags.flags['C_CONTIGUOUS'])

    def test_non_contiguous(self):
        diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1,
                                          contiguous=False)
        assert_array_equal(diags,
                           np.array([[ 4, -1, -1, -1],
                                     [ 5,  3, -1, -1],
                                     [-1,  4, -1, -1],
                                     [-1, -1, -1,  1],
                                     [-1, -1,  0,  2],
                                     [-1, -1,  1, -1]]))

        self.assertTrue(isinstance(diags.base, np.ndarray))
        self.assertFalse(diags.flags['C_CONTIGUOUS'])

    def test_boundary_node_mask_no_actives(self):
        diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1,
                                          boundary_node_mask=-2)
        assert_array_equal(diags, - 2 * np.ones((6, 4)))

    def test_boundary_node_mask(self):
        diags = sgrid.diagonal_node_array((3, 3), out_of_bounds=-1,
                                          boundary_node_mask=-2)
        assert_array_equal(diags, 
                           np.array([[-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2],
                                     [-2, -2, -2, -2], [ 8,  6,  0,  2], [-2, -2, -2, -2],
                                     [-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2]]))


class TestNeighborArray(unittest.TestCase):
    def test_default(self):
        neighbors = sgrid.neighbor_node_array((2, 3))

        BAD = sgrid.BAD_INDEX_VALUE

        assert_array_equal(
            neighbors,
            np.array([[  1,   3, BAD, BAD],
                      [  2,   4,   0, BAD],
                      [BAD,   5,   1, BAD],
                      [  4, BAD, BAD,   0],
                      [  5, BAD,   3,   1],
                      [BAD, BAD,   4,   2]]))

        self.assertTrue(neighbors.flags['C_CONTIGUOUS'])
        self.assertTrue(neighbors.base is None)

    def test_set_out_of_bounds(self):
        neighbors = sgrid.neighbor_node_array((2, 3), out_of_bounds=-1)
        assert_array_equal(neighbors,
                           np.array([[ 1,  3, -1, -1],
                                     [ 2,  4,  0, -1],
                                     [-1,  5,  1, -1],
                                     [ 4, -1, -1,  0],
                                     [ 5, -1,  3,  1],
                                     [-1, -1,  4,  2]]))

    def test_as_view(self):
        neighbors = sgrid.neighbor_node_array((2, 3), out_of_bounds=-1,
                                              contiguous=False)
        assert_array_equal(neighbors,
                           np.array([[ 1,  3, -1, -1],
                                     [ 2,  4,  0, -1],
                                     [-1,  5,  1, -1],
                                     [ 4, -1, -1,  0],
                                     [ 5, -1,  3,  1],
                                     [-1, -1,  4,  2]]))

        self.assertFalse(neighbors.flags['C_CONTIGUOUS'])
        self.assertTrue(isinstance(neighbors.base, np.ndarray))

    def test_boundary_node_mask_no_actives(self):
        neighbors = sgrid.neighbor_node_array((2, 3), out_of_bounds=-1,
                                          boundary_node_mask=-2)
        assert_array_equal(neighbors, - 2 * np.ones((6, 4)))

    def test_boundary_node_mask(self):
        neighbors = sgrid.neighbor_node_array((3, 3), out_of_bounds=-1,
                                              boundary_node_mask=-2)
        assert_array_equal(neighbors, 
                           np.array([[-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2],
                                     [-2, -2, -2, -2], [ 5,  7,  3,  1], [-2, -2, -2, -2],
                                     [-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2]]))


if __name__ == '__main__':
    unittest.main()



#! /usr/env/python
"""

Example of a simple diffusion model that uses the DiffusionComponent.

Created July 2013 GT
Last updated August 2013 GT

"""

from landlab.components.diffusion import DiffusionComponent
from landlab.grid import ModelGrid
from landlab import ModelParameterDictionary
import pylab
import numpy


def display_model(grid, elevation):
    
    # Convert z to a 2D array
    zr = grid.node_vector_to_raster(elevation, flip_vertically=True)
    
    # Create a shaded image
    pylab.close()  # clear any pre-existing plot
    im = pylab.imshow(zr, cmap=pylab.cm.RdBu)  # display a colored image
    
    # add contour lines with labels
    # add a color bar on the side
    if numpy.amax(zr)>numpy.amin(zr):
        cset = pylab.contour(zr)
        pylab.clabel(cset, inline=True, fmt='%1.1f', fontsize=10)   
        pylab.colorbar(im)
    
    # add a title
    pylab.title('Simulated topography with uplift and diffusion')

    # Display the plot
    pylab.show()
    
    
def main():
    
    # INITIALIZE
    
    # Name of parameter input file
    input_file_name = 'test_inputs_for_diffusion_model.txt'
    
    # Open input file and read run-control parameters
    mpd = ModelParameterDictionary(input_file_name)
    run_duration = mpd.get('RUN_DURATION', ptype=float)
    
    # Create and initialize a grid
    mg = ModelGrid.create_and_initialize_grid(mpd)
    
    # Create state variable (elevation)
    z = mg.create_node_dvector()
    
    # Create and initialize a diffusion component
    dc = DiffusionComponent(mg)
    dc.initialize(mpd)
    
    # RUN
    
    # Run the diffusion component until it's time for the next output
    dc.run_until(run_duration, z)
    
    # FINALIZE
    
    # Display results to screen
    display_model(mg, z)
    
 
if __name__ == "__main__":
    main()


import numpy
from collections import deque
import itertools
import scipy.sparse as sparse
import landlab.components.craters as craters #Note this brings in the data class

import time

#these ones only so we can run this module ad-hoc:
#import pylab
from pylab import plot, draw, show, contour, imshow, colorbar
#from copy import copy

#Things to add: 1. Explicit stability check. 2. Implicit handling of scenarios where kappa*dt exceeds critical step - subdivide dt automatically. 3. Don't use loops.

class perron_nl_diff_faster(object):
    '''
    This module uses Taylor Perron's implicit (2011) method to solve the nonlinear hillslope diffusion equation across a rectangular grid for a single timestep. Note it works with the mass flux implicitly, and thus does not actually calculate it. Grid must be at least 5x5.
    Built DEJH early June 2013.
    '''
    def __init__(self, grid, data, dt):
        self._delta_t = dt
        self._uplift = 0.
        self._rock_density = 2.7
        self._sed_density = 2.7
        self._kappa = 5.e-5 # ==_a
        self._S_crit = 32.*numpy.pi/180.
        self._delta_x = grid.get_grid_spacing()
        self._delta_y = self._delta_x
        self._one_over_delta_x = 1./self._delta_x
        self._one_over_delta_y = 1./self._delta_y
        self._one_over_delta_x_sqd = self._one_over_delta_x**2.
        self._one_over_delta_y_sqd = self._one_over_delta_y**2.
        self._b = 1./self._S_crit**2.
        #self._grid = grid
        #self._data = data
        
        ncols = grid.get_count_of_cols()
        nrows = grid.get_count_of_rows()
        nnodes = grid.get_count_of_all_nodes()
        #Superceded by next text block (faster)
#        self._interior_cells = list(grid.get_interior_cells())
#        _core_cells = self._interior_cells[:] #we'll thin this down into the other lists
#        _interior_corners = []
#        _interior_edges = []
#        _interior_corners.append(_core_cells.pop())
#        _interior_edges.extend(_core_cells[(4-ncols):])
#        _core_cells[(4-ncols):] = []
#        _interior_corners.append(_core_cells.pop())
#        
#        _interior_corners.append(_core_cells.pop(0))
#        _interior_edges.extend(_core_cells[:(ncols-4)])
#        _core_cells[:(ncols-4)] = []
#        _interior_corners.append(_core_cells.pop(0))
#        
#        _interior_edges.append(_core_cells.pop(0))
#        i=1
#        while 1:
#            try:
#                _interior_edges.append(_core_cells.pop(i*(ncols-4)))
#            except:
#                break
#            try:
#                _interior_edges.append(_core_cells.pop(i*(ncols-4)))
#            except:
#                break
#            else:
#                i = i+1
    
        self._interior_corners = numpy.array(ncols+1,2*ncols-2,nnodes-2*ncols+1,nnodes-ncols-2)
        _left_list = range(2*ncols+1,nnodes-2*ncols,ncols)
        _right_list = range(3*ncols-2,nnodes-2*ncols,ncols)
        _left_right = list(itertools.chain.fromiterable(itertools.izip(_left_list, _right_list)))
        self._interior_edges = numpy.array(itertools.chain(range((ncols+2),(2*ncols-2)),_left_right,range(nnodes-2*ncols+2,nnodes-ncols-2)))
        self._core_cells = numpy.zeros(grid.get_count_of_interior_nodes()-2*ncols-4-2*(nrows-4),dtype=int)
        for i in range(nrows-4):
            _core_cells[i*(ncols-4):(i+1)*(ncols-4)] = range((2+i)*ncols+2,(3+i)*ncols-2) #fill IDs into core cell matrix
        
        #self._core_cells = numpy.array(_core_cells)
        #self._interior_edges = numpy.array(_interior_edges) #order is [ncols-4 of TOP]+[ncols-4 of BOTTOM]+[(nrows-4)*(LEFT,RIGHT) pairs]
        #self._interior_corners = numpy.array(_interior_corners) #order is topright,topleft,bottomleft,bottomright
        #print _interior_corners
        #print _interior_edges
        #print _core_cells
        #setup is giving the correct interior cell lists here


    def set_variables(self, grid, data):
        '''
        This function sets the variables needed for update().
        '''
        #self._grid = grid
        #self._data = data
        n_interior_cells = grid.get_count_of_interior_cells()
        _operating_matrix = sparse.lil_matrix((n_interior_cells, n_interior_cells), dtype=float)
        #_interior_elevs = [-1] * n_interior_cells

        #Initialize the local builder lists
        _mat_RHS = numpy.zeros(n_interior_cells)
        
        self.set_variables_for_core_cells(grid, data, _operating_matrix, _mat_RHS)
        self.set_variables_for_interior_edges(grid, data, _operating_matrix, _mat_RHS)
        self.set_variables_for_interior_corners(grid, data, _operating_matrix, _mat_RHS)
    
        self._operating_matrix = _operating_matrix.tocsr()
        self._mat_RHS = _mat_RHS


    def set_variables_for_core_cells(self, grid, data, _operating_matrix, _mat_RHS):
    
        elev = data.elev
        ncols = grid.get_count_of_cols()
        _delta_t = self._delta_t
        _one_over_delta_x = self._one_over_delta_x
        _one_over_delta_x_sqd = self._one_over_delta_x_sqd
        _one_over_delta_y = self._one_over_delta_y
        _one_over_delta_y_sqd = self._one_over_delta_y_sqd
        _kappa = self._kappa
        _b = self._b
        _S_crit = self._S_crit
        count = 0
        interior_grid_width = ncols-2
        core_cell_width = ncols-4
        for i in self._core_cells:
            n = (count//core_cell_width+1)*interior_grid_width + (count%core_cell_width) + 1 #This is the ID within the interior grid
            #n_test = self._interior_cells.index(i)
            #assert n == n_test
            cell_neighbors = grid.get_neighbor_list(i)
            cell_diagonals = grid.get_diagonal_list(i)
            _z_x = (data.elev[cell_neighbors[0]]-data.elev[cell_neighbors[2]])*0.5*_one_over_delta_x
            _z_y = (data.elev[cell_neighbors[1]]-data.elev[cell_neighbors[3]])*0.5*_one_over_delta_y
            _z_xx = (data.elev[cell_neighbors[0]]-2.*data.elev[i]+data.elev[cell_neighbors[2]])*_one_over_delta_x_sqd
            _z_yy = (data.elev[cell_neighbors[1]]-2.*data.elev[i]+data.elev[cell_neighbors[3]])*_one_over_delta_y_sqd
            _z_xy = (data.elev[cell_diagonals[0]] - data.elev[cell_diagonals[1]] - data.elev[cell_diagonals[3]] + data.elev[cell_diagonals[2]])*0.25*_one_over_delta_x*_one_over_delta_y
            _d = 1./(1.-_b*(_z_x*_z_x+_z_y*_z_y))
            
            #assert type(_z_x) is numpy.float64
            #assert type(_z_y) is numpy.float64
            #assert type(_z_xx) is numpy.float64
            #assert type(_z_yy) is numpy.float64
            #assert type(_z_xy) is numpy.float64
            
            _abd_sqd = _kappa*_b*_d*_d
            _F_ij = -2.*_kappa*_d*(_one_over_delta_x_sqd+_one_over_delta_y_sqd) - 4.*_abd_sqd*(_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_y*_one_over_delta_y_sqd)
            _F_ijminus1 = _kappa*_d*_one_over_delta_x_sqd - _abd_sqd*_z_x*(_z_xx+_z_yy)*_one_over_delta_x - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*_one_over_delta_x - 2.*_abd_sqd*(_z_x*_z_xx*_one_over_delta_x-_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_xy*_one_over_delta_x)
            _F_ijplus1 = _kappa*_d*_one_over_delta_x_sqd + _abd_sqd*_z_x*(_z_xx+_z_yy)*_one_over_delta_x + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*_one_over_delta_x + 2.*_abd_sqd*(_z_x*_z_xx*_one_over_delta_x+_z_x*_z_x*_one_over_delta_x_sqd+_z_y*_z_xy*_one_over_delta_x)
            _F_iminus1j = _kappa*_d*_one_over_delta_y_sqd - _abd_sqd*_z_y*(_z_xx+_z_yy)*_one_over_delta_y - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*_one_over_delta_y - 2.*_abd_sqd*(_z_y*_z_yy*_one_over_delta_y-_z_y*_z_y*_one_over_delta_y_sqd+_z_x*_z_xy*_one_over_delta_y)
            _F_iplus1j = _kappa*_d*_one_over_delta_y_sqd + _abd_sqd*_z_y*(_z_xx+_z_yy)*_one_over_delta_y + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*_one_over_delta_y + 2.*_abd_sqd*(_z_y*_z_yy*_one_over_delta_y+_z_y*_z_y*_one_over_delta_y_sqd+_z_x*_z_xy*_one_over_delta_y)
            _F_iplus1jplus1 = _abd_sqd*_z_x*_z_y*_one_over_delta_x*_one_over_delta_y
            _F_iminus1jminus1 = _F_iplus1jplus1
            _F_iplus1jminus1 = -_F_iplus1jplus1
            _F_iminus1jplus1 = _F_iplus1jminus1
            
            #assert type(_F_ij) is numpy.float64
            #assert type(_F_ijminus1) is numpy.float64
            #assert type(_F_ijplus1) is numpy.float64
            #assert type(_F_iminus1j) is numpy.float64
            #assert type(_F_iplus1j) is numpy.float64
            #assert type(_F_iplus1jplus1) is numpy.float64

            #RHS of equ 6 (see para [20])
            _func_on_z = self._rock_density/self._sed_density*self._uplift + _kappa*((_z_xx+_z_yy)/(1.-(_z_x*_z_x+_z_y*_z_y)/_S_crit*_S_crit) + 2.*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)/(_S_crit*_S_crit*(1.-(_z_x*_z_x+_z_y*_z_y)/_S_crit*_S_crit)**2.))
                        
            #assert type(_func_on_z) is numpy.float64

            _mat_RHS[n] = _mat_RHS[n] + data.elev[i] + _delta_t*(_func_on_z - (_F_ij*data.elev[i]+_F_ijminus1*data.elev[cell_neighbors[2]]+_F_ijplus1*data.elev[cell_neighbors[0]]+_F_iminus1j*data.elev[cell_neighbors[3]]+_F_iplus1j*data.elev[cell_neighbors[1]]+_F_iminus1jminus1*data.elev[cell_diagonals[2]]+_F_iplus1jplus1*data.elev[cell_diagonals[0]]+_F_iplus1jminus1*data.elev[cell_diagonals[1]]+_F_iminus1jplus1*data.elev[cell_diagonals[3]]))
            
            #build the operating matrix. No logic operations needed as these are all core cells
            _operating_matrix[n,(n-1):(n+2)] += [-_delta_t*_F_ijminus1, 1.-_delta_t*_F_ij, -_delta_t*_F_ijplus1]
            #_operating_matrix[n,n] = _operating_matrix[n,n]+1.-_delta_t*_F_ij
            #_operating_matrix[n,n-1] = _operating_matrix[n,n-1]-_delta_t*_F_ijminus1
            #_operating_matrix[n,n+1] = _operating_matrix[n,n+1]-_delta_t*_F_ijplus1
            _operating_matrix[n,(n-ncols+1):(n-ncols+4)] -= [_F_iminus1jminus1, _F_iminus1j, _F_iminus1jplus1]*_delta_t
            #_operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-_delta_t*_F_iminus1j
            #_operating_matrix[n,n-ncols+1] = _operating_matrix[n,n-ncols+1]-_delta_t*_F_iminus1jminus1
            #_operating_matrix[n,n-ncols+3] = _operating_matrix[n,n-ncols+3]-_delta_t*_F_iminus1jplus1
            _operating_matrix[n,(n+ncols-3):(n+ncols)] -= [_F_iplus1jminus1, _F_iplus1j, _F_iplus1jplus1]*_delta_t
            #_operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-_delta_t*_F_iplus1j
            #_operating_matrix[n,n+ncols-1] = _operating_matrix[n,n+ncols-1]-_delta_t*_F_iplus1jplus1
            #_operating_matrix[n,n+ncols-3] = _operating_matrix[n,n+ncols-3]-_delta_t*_F_iplus1jminus1
        
            count = count + 1


    def set_variables_for_interior_corners(self, grid, data, _operating_matrix, _mat_RHS):
        ncols = grid.get_count_of_cols()
        count = 0
        corner_ids = [-1,-(ncols-2),0,ncols-3] #topright,topleft,bottomleft,bottomright. This is in the ID reference frame of the INTERIOR GRID
        for i in self._interior_corners:
            #print i
            n = corner_ids[count]
            cell_neighbors = grid.get_neighbor_list(i)
            cell_diagonals = grid.get_diagonal_list(i)
            _z_x = (data.elev[cell_neighbors[0]]-data.elev[cell_neighbors[2]])*0.5*self._one_over_delta_x
            _z_y = (data.elev[cell_neighbors[1]]-data.elev[cell_neighbors[3]])*0.5*self._one_over_delta_y
            _z_xx = (data.elev[cell_neighbors[0]]-2.*data.elev[i]+data.elev[cell_neighbors[2]])*self._one_over_delta_x_sqd
            _z_yy = (data.elev[cell_neighbors[1]]-2.*data.elev[i]+data.elev[cell_neighbors[3]])*self._one_over_delta_y_sqd
            _z_xy = (data.elev[cell_diagonals[0]] - data.elev[cell_diagonals[1]] - data.elev[cell_diagonals[3]] + data.elev[cell_diagonals[2]])*0.25*self._one_over_delta_x*self._one_over_delta_y
            _d = 1./(1.-self._b*(_z_x**2.+_z_y**2.))
            
            #assert type(_z_x) is numpy.float64
            #assert type(_z_y) is numpy.float64
            #assert type(_z_xx) is numpy.float64
            #assert type(_z_yy) is numpy.float64
            #assert type(_z_xy) is numpy.float64
            
            _abd_sqd = self._kappa*self._b*_d**2.
            _F_ij = -2.*self._kappa*_d*(self._one_over_delta_x_sqd+self._one_over_delta_y_sqd) - 4.*_abd_sqd*(_z_x**2.*self._one_over_delta_x_sqd+_z_y**2.*self._one_over_delta_y_sqd)
            _F_ijminus1 = self._kappa*_d*self._one_over_delta_x_sqd - _abd_sqd*_z_x*(_z_xx+_z_yy)*self._one_over_delta_x - 4.*_abd_sqd*self._b*_d*(_z_x**2.*_z_xx+_z_y**2.*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*self._one_over_delta_x - 2.*_abd_sqd*(_z_x*_z_xx*self._one_over_delta_x-_z_x**2.*self._one_over_delta_x_sqd+_z_y*_z_xy*self._one_over_delta_x)
            _F_ijplus1 = self._kappa*_d*self._one_over_delta_x_sqd + _abd_sqd*_z_x*(_z_xx+_z_yy)*self._one_over_delta_x + 4.*_abd_sqd*self._b*_d*(_z_x**2.*_z_xx+_z_y**2.*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*self._one_over_delta_x + 2.*_abd_sqd*(_z_x*_z_xx*self._one_over_delta_x+_z_x**2.*self._one_over_delta_x_sqd+_z_y*_z_xy*self._one_over_delta_x)
            _F_iminus1j = self._kappa*_d*self._one_over_delta_y_sqd - _abd_sqd*_z_y*(_z_xx+_z_yy)*self._one_over_delta_y - 4.*_abd_sqd*self._b*_d*(_z_x**2.*_z_xx+_z_y**2.*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*self._one_over_delta_y - 2.*_abd_sqd*(_z_y*_z_yy*self._one_over_delta_y-_z_y**2.*self._one_over_delta_y_sqd+_z_x*_z_xy*self._one_over_delta_y)
            _F_iplus1j = self._kappa*_d*self._one_over_delta_y_sqd + _abd_sqd*_z_y*(_z_xx+_z_yy)*self._one_over_delta_y + 4.*_abd_sqd*self._b*_d*(_z_x**2.*_z_xx+_z_y**2.*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*self._one_over_delta_y + 2.*_abd_sqd*(_z_y*_z_yy*self._one_over_delta_y+_z_y**2.*self._one_over_delta_y_sqd+_z_x*_z_xy*self._one_over_delta_y)
            _F_iplus1jplus1 = _abd_sqd*_z_x*_z_y*self._one_over_delta_x*self._one_over_delta_y
            _F_iminus1jminus1 = _F_iplus1jplus1
            _F_iplus1jminus1 = -_F_iplus1jplus1
            _F_iminus1jplus1 = _F_iplus1jminus1
            
            #assert type(_F_ij) is numpy.float64
            #assert type(_F_ijminus1) is numpy.float64
            #assert type(_F_ijplus1) is numpy.float64
            #assert type(_F_iminus1j) is numpy.float64
            #assert type(_F_iplus1j) is numpy.float64
            #assert type(_F_iplus1jplus1) is numpy.float64

            #RHS of equ 6 (see para [20])
            _func_on_z = self._rock_density/self._sed_density*self._uplift + self._kappa*((_z_xx+_z_yy)/(1.-(_z_x**2.+_z_y**2.)/self._S_crit**2.) + 2.*(_z_x**2.*_z_xx+_z_y**2.*_z_yy+2.*_z_x*_z_y*_z_xy)/(self._S_crit**2.*(1.-(_z_x**2.+_z_y**2.)/self._S_crit**2.)**2.))
                        
            #assert type(_func_on_z) is numpy.float64

            _mat_RHS[n] = _mat_RHS[n] + data.elev[i] + self._delta_t*(_func_on_z - (_F_ij*data.elev[i]+_F_ijminus1*data.elev[cell_neighbors[2]]+_F_ijplus1*data.elev[cell_neighbors[0]]+_F_iminus1j*data.elev[cell_neighbors[3]]+_F_iplus1j*data.elev[cell_neighbors[1]]+_F_iminus1jminus1*data.elev[cell_diagonals[2]]+_F_iplus1jplus1*data.elev[cell_diagonals[0]]+_F_iplus1jminus1*data.elev[cell_diagonals[1]]+_F_iminus1jplus1*data.elev[cell_diagonals[3]]))
            
            #build the operating matrix
            _operating_matrix[n,n] = _operating_matrix[n,n]+1.-self._delta_t*_F_ij
            if count==0: #topright
                _operating_matrix[n,n-1] = _operating_matrix[n,n-1]-self._delta_t*_F_ijminus1 #left
                _operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-self._delta_t*_F_iminus1j #below
                _operating_matrix[n,n-ncols+1] = _operating_matrix[n,n-ncols+1]-self._delta_t*_F_iminus1jminus1 #leftbelow
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols-1, _F_iplus1jminus1) #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, ncols, _F_iplus1j) #above
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols+1, _F_iplus1jplus1) #aboveright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, 1, _F_ijplus1) #right
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols+1, _F_iminus1jplus1) #belowright
            if count==1: #topleft
                _operating_matrix[n,n+1] = _operating_matrix[n,n+1]-self._delta_t*_F_ijplus1 #right
                _operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-self._delta_t*_F_iminus1j #below
                _operating_matrix[n,n-ncols+3] = _operating_matrix[n,n-ncols+3]-self._delta_t*_F_iminus1jplus1 #belowright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols-1, _F_iplus1jminus1) #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, ncols, _F_iplus1j) #above
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols+1, _F_iplus1jplus1) #aboveright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -1, _F_ijminus1) #left
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols-1, _F_iminus1jminus1) #belowleft
            if count==2: #bottomleft
                _operating_matrix[n,n+1] = _operating_matrix[n,n+1]-self._delta_t*_F_ijplus1 #right
                _operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-self._delta_t*_F_iplus1j #above
                _operating_matrix[n,n+ncols-1] = _operating_matrix[n,n+ncols-1]-self._delta_t*_F_iplus1jplus1 #aboveright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols-1, _F_iplus1jminus1) #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -1, _F_ijminus1) #left
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols-1, _F_iminus1jminus1) #belowleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols, _F_iminus1j) #below
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols+1, _F_iminus1jplus1) #belowright
            if count==3: #bottomright
                _operating_matrix[n,n-1] = _operating_matrix[n,n-1]-self._delta_t*_F_ijminus1 #left
                _operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-self._delta_t*_F_iplus1j #above
                _operating_matrix[n,n+ncols-3] = _operating_matrix[n,n+ncols-3]-self._delta_t*_F_iplus1jminus1 #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols+1, _F_iplus1jplus1) #aboveright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, 1, _F_ijplus1) #right
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols+1, _F_iminus1jplus1) #belowright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols, _F_iminus1j) #below
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols-1, _F_iminus1jminus1) #belowleft
            count = count+1
                

    def set_variables_for_interior_edges(self, grid, data, _operating_matrix, _mat_RHS):
        count = 0
        side_IDs = []
        nrows = grid.get_count_of_rows()
        ncols = grid.get_count_of_cols()
        for i in range(nrows-4):
            side_IDs.extend([(1+i)*(ncols-2), (2+i)*(ncols-2)-1])
        edge_ids = range(-(ncols-3),-1)+range(1,(ncols-3))+side_IDs #order is [ncols-4 of TOP]+[ncols-4 of BOTTOM]+[(nrows-4)*(LEFT,RIGHT) pairs]; ID to the INTERIOR GRID
        #print edge_ids
        for i in self._interior_edges:
            #print '---'
            #print count
            #print i
            n = edge_ids[count]
            #print n
            cell_neighbors = grid.get_neighbor_list(i)
            cell_diagonals = grid.get_diagonal_list(i)
            _z_x = (data.elev[cell_neighbors[0]]-data.elev[cell_neighbors[2]])*0.5*self._one_over_delta_x
            _z_y = (data.elev[cell_neighbors[1]]-data.elev[cell_neighbors[3]])*0.5*self._one_over_delta_y
            _z_xx = (data.elev[cell_neighbors[0]]-2.*data.elev[i]+data.elev[cell_neighbors[2]])*self._one_over_delta_x_sqd
            _z_yy = (data.elev[cell_neighbors[1]]-2.*data.elev[i]+data.elev[cell_neighbors[3]])*self._one_over_delta_y_sqd
            _z_xy = (data.elev[cell_diagonals[0]] - data.elev[cell_diagonals[1]] - data.elev[cell_diagonals[3]] + data.elev[cell_diagonals[2]])*0.25*self._one_over_delta_x*self._one_over_delta_y
            _d = 1./(1.-self._b*(_z_x*_z_x+_z_y*_z_y))
            
            #assert type(_z_x) is numpy.float64
            #assert type(_z_y) is numpy.float64
            #assert type(_z_xx) is numpy.float64
            #assert type(_z_yy) is numpy.float64
            #assert type(_z_xy) is numpy.float64
            
            _abd_sqd = self._kappa*self._b*_d*_d
            _F_ij = -2.*self._kappa*_d*(self._one_over_delta_x_sqd+self._one_over_delta_y_sqd) - 4.*_abd_sqd*(_z_x*_z_x*self._one_over_delta_x_sqd+_z_y*_z_y*self._one_over_delta_y_sqd)
            _F_ijminus1 = self._kappa*_d*self._one_over_delta_x_sqd - _abd_sqd*_z_x*(_z_xx+_z_yy)*self._one_over_delta_x - 4.*_abd_sqd*self._b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*self._one_over_delta_x - 2.*_abd_sqd*(_z_x*_z_xx*self._one_over_delta_x-_z_x*_z_x*self._one_over_delta_x_sqd+_z_y*_z_xy*self._one_over_delta_x)
            _F_ijplus1 = self._kappa*_d*self._one_over_delta_x_sqd + _abd_sqd*_z_x*(_z_xx+_z_yy)*self._one_over_delta_x + 4.*_abd_sqd*self._b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_x*self._one_over_delta_x + 2.*_abd_sqd*(_z_x*_z_xx*self._one_over_delta_x+_z_x*_z_x*self._one_over_delta_x_sqd+_z_y*_z_xy*self._one_over_delta_x)
            _F_iminus1j = self._kappa*_d*self._one_over_delta_y_sqd - _abd_sqd*_z_y*(_z_xx+_z_yy)*self._one_over_delta_y - 4.*_abd_sqd*self._b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*self._one_over_delta_y - 2.*_abd_sqd*(_z_y*_z_yy*self._one_over_delta_y-_z_y*_z_y*self._one_over_delta_y_sqd+_z_x*_z_xy*self._one_over_delta_y)
            _F_iplus1j = self._kappa*_d*self._one_over_delta_y_sqd + _abd_sqd*_z_y*(_z_xx+_z_yy)*self._one_over_delta_y + 4.*_abd_sqd*self._b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)*_z_y*self._one_over_delta_y + 2.*_abd_sqd*(_z_y*_z_yy*self._one_over_delta_y+_z_y*_z_y*self._one_over_delta_y_sqd+_z_x*_z_xy*self._one_over_delta_y)
            _F_iplus1jplus1 = _abd_sqd*_z_x*_z_y*self._one_over_delta_x*self._one_over_delta_y
            _F_iminus1jminus1 = _F_iplus1jplus1
            _F_iplus1jminus1 = -_F_iplus1jplus1
            _F_iminus1jplus1 = _F_iplus1jminus1
            
            #assert type(_F_ij) is numpy.float64
            #assert type(_F_ijminus1) is numpy.float64
            #assert type(_F_ijplus1) is numpy.float64
            #assert type(_F_iminus1j) is numpy.float64
            #assert type(_F_iplus1j) is numpy.float64
            #assert type(_F_iplus1jplus1) is numpy.float64

            #RHS of equ 6 (see para [20])
            _func_on_z = self._rock_density/self._sed_density*self._uplift + self._kappa*((_z_xx+_z_yy)/(1.-(_z_x*_z_x+_z_y*_z_y)/self._S_crit*self._S_crit) + 2.*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy)/(self._S_crit*self._S_crit*(1.-(_z_x*_z_x+_z_y*_z_y)/self._S_crit*self._S_crit)**2.))
                        
            #assert type(_func_on_z) is numpy.float64

            _mat_RHS[n] = _mat_RHS[n] + data.elev[i] + self._delta_t*(_func_on_z - (_F_ij*data.elev[i]+_F_ijminus1*data.elev[cell_neighbors[2]]+_F_ijplus1*data.elev[cell_neighbors[0]]+_F_iminus1j*data.elev[cell_neighbors[3]]+_F_iplus1j*data.elev[cell_neighbors[1]]+_F_iminus1jminus1*data.elev[cell_diagonals[2]]+_F_iplus1jplus1*data.elev[cell_diagonals[0]]+_F_iplus1jminus1*data.elev[cell_diagonals[1]]+_F_iminus1jplus1*data.elev[cell_diagonals[3]]))
            
            #build the operating matrix
            _operating_matrix[n,n] = _operating_matrix[n,n]+1.-self._delta_t*_F_ij
            if count<ncols-4: #top
                _operating_matrix[n,n-1] = _operating_matrix[n,n-1]-self._delta_t*_F_ijminus1 #left
                _operating_matrix[n,n+1] = _operating_matrix[n,n+1]-self._delta_t*_F_ijplus1 #right
                _operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-self._delta_t*_F_iminus1j #below
                _operating_matrix[n,n-ncols+1] = _operating_matrix[n,n-ncols+1]-self._delta_t*_F_iminus1jminus1 #leftbelow
                _operating_matrix[n,n-ncols+3] = _operating_matrix[n,n-ncols+3]-self._delta_t*_F_iminus1jplus1 #belowright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols-1, _F_iplus1jminus1) #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, ncols, _F_iplus1j) #above
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols+1, _F_iplus1jplus1) #aboveright
            elif count<2*(ncols-4): #bottom
                _operating_matrix[n,n-1] = _operating_matrix[n,n-1]-self._delta_t*_F_ijminus1 #left
                _operating_matrix[n,n+1] = _operating_matrix[n,n+1]-self._delta_t*_F_ijplus1 #right
                _operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-self._delta_t*_F_iplus1j #above
                _operating_matrix[n,n+ncols-1] = _operating_matrix[n,n+ncols-1]-self._delta_t*_F_iplus1jplus1 #aboveright
                _operating_matrix[n,n+ncols-3] = _operating_matrix[n,n+ncols-3]-self._delta_t*_F_iplus1jminus1 #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols-1, _F_iminus1jminus1) #belowleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols, _F_iminus1j) #below
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols+1, _F_iminus1jplus1) #belowright
            elif (count-2*(ncols-4))%2: #right
                _operating_matrix[n,n-1] = _operating_matrix[n,n-1]-self._delta_t*_F_ijminus1 #left
                _operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-self._delta_t*_F_iplus1j #above
                _operating_matrix[n,n+ncols-3] = _operating_matrix[n,n+ncols-3]-self._delta_t*_F_iplus1jminus1 #aboveleft
                _operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-self._delta_t*_F_iminus1j #below
                _operating_matrix[n,n-ncols+1] = _operating_matrix[n,n-ncols+1]-self._delta_t*_F_iminus1jminus1 #leftbelow
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols+1, _F_iplus1jplus1) #aboveright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, 1, _F_ijplus1) #right
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols+1, _F_iminus1jplus1) #belowright
            else: #left
                _operating_matrix[n,n+1] = _operating_matrix[n,n+1]-self._delta_t*_F_ijplus1 #right
                _operating_matrix[n,n+ncols-2] = _operating_matrix[n,n+ncols-2]-self._delta_t*_F_iplus1j #above
                _operating_matrix[n,n+ncols-1] = _operating_matrix[n,n+ncols-1]-self._delta_t*_F_iplus1jplus1 #aboveright
                _operating_matrix[n,n-ncols+2] = _operating_matrix[n,n-ncols+2]-self._delta_t*_F_iminus1j #below
                _operating_matrix[n,n-ncols+3] = _operating_matrix[n,n-ncols+3]-self._delta_t*_F_iminus1jplus1 #belowright
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, +ncols-1, _F_iplus1jminus1) #aboveleft
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -1, _F_ijminus1) #left
                self.set_bc_cell(grid, data, _operating_matrix, _mat_RHS, n, -ncols-1, _F_iminus1jminus1) #belowleft

            count = count+1
        
    #This needs updating to reflect Greg's new BC handling in model_grid
    def set_bc_cell(self, grid, data, _operating_matrix, _mat_RHS, n, modulator, _F_value):
        bc_pos = grid.boundary_cells.tolist().index(self._interior_cells[n]+modulator)
        if grid.default_bc.boundary_code[bc_pos] == 3: #tracks cell
            #Find the cell it's linked to in *this* internal-cells-only matrix:
            l = self._interior_cells.tolist().index(grid.default_bc.tracks_cell[bc_pos])
            _operating_matrix[n,l] = _operating_matrix[n,l]-self._delta_t*_F_value
        elif grid.default_bc.boundary_code[bc_pos] == 2: #fixed gradient
            #Which cell is it linked to?
            l = self._interior_cells.tolist().index(grid.default_bc.tracks_cell[bc_pos])
            #One part goes in this cell...
            _operating_matrix[n,l] = _operating_matrix[n,l]-self._delta_t*_F_value
            #And one part goes on the RHS
            _mat_RHS[n] = _mat_RHS[n]+self._delta_t*_F_value*self._delta_x*grid.default_bc.boundary_gradient[bc_pos]
        elif grid.default_bc.boundary_code[bc_pos] == 1: #fixed value
            #Goes to the RHS
            _mat_RHS[n] = _mat_RHS[n]+self._delta_t*_F_value*data.elev[self._interior_cells[n]+modulator]
        else:
            try:
                raise IndexError
            except IndexError:
                print "Boundary code was not of an expected value!"
                raise


    def update(self, grid, data_in):
        #Initialize the variables for the step:
        self.set_variables(grid, data_in)
        #Solve interior of grid:
        _interior_elevs = sparse.linalg.spsolve(self._operating_matrix, self._mat_RHS)
        #print _interior_elevs.shape
        #this fn solves Ax=B for x
        
        #Load the new elevs into the actual elev data
        j=0 #counter for location in the matrix _interior_elevs
        ncells = grid.get_number_of_nodes()
        for i in range(ncells):
            if grid.is_interior(i):
                data_in.elev[i] = _interior_elevs[j]
                j = j+1
            else:
                bc_pos = grid.boundary_cells.tolist().index(i)
                if grid.default_bc.boundary_code[bc_pos] == 3:
                    data_in.elev[i] = _interior_elevs[0,self._interior_cells.tolist().index(grid.default_bc.tracks_cell[bc_pos])]
                elif grid.default_bc.boundary_code[bc_pos] == 2:
                    data_in.elev[i] = _interior_elevs[0,self._interior_cells.tolist().index(grid.default_bc.tracks_cell[bc_pos])] + grid.default_bc.gradient[bc_pos]*self._delta_x
                elif grid.default_bc.boundary_code[bc_pos] == 1:
                    continue #The old value of elev is still correct
                else:
                    try:
                        raise IndexError
                    except IndexError:
                        print "Boundary code was not of an expected value!"
                        raise


def main():

    dt = 1.
    nt = 500
    cr, mg, vectors = craters.dig_one_crater(120,120, 0.025, 0.5, 0.5, 1.)
    start_time = time.time()
    diff = perron_nl_diff_faster(mg, vectors, dt)
    for i in range (nt):
        diff.update(mg, vectors)
        print "Loop number ", i+1, " completed!"

    #Finalize
    elev_raster = mg.cell_vector_to_raster(vectors.elev)

    end_time = time.time()
    print('Elapsed time was %g seconds' % (end_time - start_time))

    #contour(elev_raster)
    flipped_elev_raster = numpy.empty_like(elev_raster)
    for i in range(0,mg.nrows):
        flipped_elev_raster[i,:] = elev_raster[(mg.nrows-i-1),:]
    imshow(flipped_elev_raster)
    colorbar()
    show()
    vectors.viewing_raster = flipped_elev_raster


if __name__=='__main__':
    main()


#! /usr/bin/env python
"""
Read data from a NetCDF file into a RasterModelGrid.
"""

try:
    import netCDF4 as nc
except ImportError:
    import warnings
    warnings.warn('Unable to input netCDF4.', ImportWarning)

import os
import types
import re

import numpy as np

from landlab import RasterModelGrid
from landlab.io.netcdf.errors import NotRasterGridError
from landlab.io.netcdf._constants import (_AXIS_DIMENSION_NAMES,
                                          _AXIS_COORDINATE_NAMES,
                                          _COORDINATE_NAMES)


def _read_netcdf_grid_shape(root):
    shape = []
    for axis_name in _AXIS_DIMENSION_NAMES:
        try:
            shape.append(len(root.dimensions[axis_name]))
        except KeyError:
            pass
    return shape


def _read_netcdf_coordinate_values(root):
    values = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            values.append(root.variables[coordinate_name][:])
        except KeyError:
            pass
    return values


def _read_netcdf_coordinate_units(root):
    units = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            units.append(root.variables[coordinate_name].units)
        except KeyError:
            pass
    return units


def _read_netcdf_structured_grid(root):
    shape = _read_netcdf_grid_shape(root)
    coordinates = _read_netcdf_coordinate_values(root)
    units = _read_netcdf_coordinate_units(root)

    for coordinate in coordinates:
        coordinate.shape = shape

    return coordinates


def _read_netcdf_structured_data(root):
    fields = dict()
    for (name, var) in root.variables.items():
        if not name in _COORDINATE_NAMES:
            fields[name] = var[:]

    return fields


def _get_raster_spacing(coords):
    spacing = np.empty(len(coords), dtype=np.float64)

    for (axis, coord) in enumerate(coords):
        coord_spacing = np.diff(coord, axis=axis)
        try:
            assert(np.all(coord_spacing == coord_spacing.flat[0]))
        except AssertionError:
            raise NotRasterGridError()
        spacing[axis] = coord_spacing.flat[0]

    try:
        assert(np.all(spacing == spacing[0]))
    except AssertionError:
        raise NotRasterGridError()
    else:
        return spacing[0]


def read_netcdf(nc_file, reshape=False, just_grid=False):
    root = nc.Dataset(nc_file, 'r', format='NETCDF4')

    #(node_coords, units) = _read_netcdf_structured_grid(root)
    node_coords = _read_netcdf_structured_grid(root)

    try:
        assert(len(node_coords) == 2)
    except AssertionError:
        raise

    try:
        spacing = _get_raster_spacing(node_coords)
    except NotRasterGridError:
        print node_coords
        raise 

    shape = node_coords[0].shape

    grid = RasterModelGrid(num_rows=shape[0], num_cols=shape[1],
                           dx=spacing)

    if not just_grid:
        fields = _read_netcdf_structured_data(root)
    else:
        fields = dict()

    root.close()

    return (grid, fields)



#! /usr/bin/env python

import types
import inspect

from landlab.field import ScalarDataFields


class ModelDataFields(dict):
    def __init__(self):
        super(ModelDataFields, self).__init__()

    @property
    def groups(self):
        return set(self.keys())

    def new_field_location(self, group, size):
        if group not in self:
            self[group] = ScalarDataFields(size)
            setattr(self, 'at_' + group, self[group])
        else:
            raise ValueError('ModelDataField already contains %s' % group)

    def field_values(self, group, field):
        return self[group][field]

    def field_units(self, group, field):
        return self[group].units[field]


def get_method_from_class(clazz, method_name):
    method = getattr(clazz, method_name)
    if inspect.ismethod(method):
        return method
    else:
        raise TypeError('%s is not a bound method' % method_name)


def is_regular_method_name(name):
    return not (name.startswith('__') and name.endswith('__'))


def regular_method_names(clazz):
    methods = set()
    for (name, _) in inspect.getmembers(clazz, inspect.ismethod):
        if is_regular_method_name(name):
            methods.add(name)
    return methods


def prepend_arg_list_with_dict_value(destination_class, method):
    def func_with_dict_value_first(self, key, *args, **kwds):
        return method(self[key], *args, **kwds)
    func_with_dict_value_first.__doc__ = method.__doc__
    func_with_dict_value_first.__name__ = method.__name__
    return types.MethodType(func_with_dict_value_first, None,
                            destination_class)


def add_methods_to_grouped_fields_class():
    for name in (regular_method_names(ScalarDataFields) -
                 regular_method_names(ModelDataFields)):
        method = get_method_from_class(ScalarDataFields, name)
        new_method = prepend_arg_list_with_dict_value(ModelDataFields,
                                                      method)
        setattr(ModelDataFields, name, new_method)


add_methods_to_grouped_fields_class()



#! /usr/bin/env python

import pylab

from landlab.components.flexure import Flexure


def main():

    shape = (100, 100)
    spacing = (10e3, 10e3)

    flex = Flexure(shape, spacing, (0., 0.))

    load = flex['lithosphere__overlying_pressure']
    load[shape[0]/2, shape[1]/2] = 1e9

    dz = flex['lithosphere__elevation']

    flex.update()

    dz.shape = shape

    pylab.imshow(dz)
    pylab.colorbar()
    #pylab.plot(dz[shape[0]/2,:])
    pylab.show()


if __name__ == '__main__':
    main()



#! /usr/bin/env python

import numpy as np
import matplotlib.pyplot as plt


def imshow_grid(grid, values, var_name=None, var_units=None,
                grid_units=(None, None), symmetric_cbar=False):
    data = values.view()
    data.shape = grid.shape

    y = np.linspace(0, grid.get_grid_ydimension(), data.shape[0])
    x = np.linspace(0, grid.get_grid_xdimension(), data.shape[1])

    if symmetric_cbar:
        (var_min, var_max) = (data.min(), data.max())
        limit = max(abs(var_min), abs(var_max))
        limits = (-limit, limit)
    else:
        limits = (None, None)

    plt.pcolormesh(y, x, data, vmin=limits[0], vmax=limits[1])
    plt.colorbar()

    plt.xlabel('X (%s)' % grid_units[1])
    plt.ylabel('Y (%s)' % grid_units[0])

    if var_name is not None:
        plt.title('%s (%s)' % (var_name, var_units))

    plt.show()


def imshow_field(field, name, **kwds):
    imshow_grid(field, field[name], var_name=name, var_units=field.units[name],
                **kwds)



import landlab.components.craters as craters
from landlab import RasterModelGrid
import numpy
from pylab import show, imshow, colorbar, plot
from copy import copy

craters=reload(craters)

#craters_data = craters.one_crater_then_degrade()

class data(object):
    '''
        This is where all the whole-grid data lives, as arrays over the various elements of the grid.
        '''
    #Data goes here!!!
    def __init__(self, grid):
        self.elev = grid.create_node_dvector() #some data
        self.flag_already_in_the_list = grid.create_node_dvector()
        self.craters_over_max_radius_not_plotted = grid.create_node_dvector()
        self.impact_sequence = []

#def dig_one_crater_driver(nr, nc, dx, rel_x, rel_y, radius=numpy.nan, angle=numpy.nan):
#    '''
#        This is an ad-hoc script to dig one crater.
#        '''
#    
#    #Setup the grid & impactor object
#    mg = RasterModelGrid()
#    mg.initialize(nr, nc, dx)
#    vectors = data(mg)
#    vectors.elev[:] = 0.
#    cr = craters.impactor()
#
#    #Dig the crater
#    cr.excavate_a_crater_optimized(mg, vectors, forced_radius=radius, forced_angle=angle, forced_pos=[rel_x, rel_y])
#
#    #Finalize
#    elev_raster = mg.node_vector_to_raster(vectors.elev, flip_vertically=True)
#    return elev_raster


#elev_raster = dig_one_crater_driver(120, 120, 0.025, 0.5, 0.5, 0.75, 90.)
#imshow(elev_raster)
#colorbar()
#show()

##This code snippet to produce a landscape w. 500000 impacts, but where the first half of the impacts are 10x bigger on average.
#mg = RasterModelGrid()
#mg.initialize(1200, 1200, 0.0025)
#vectors = data(mg)
#vectors.elev[:] = 0.
#
#dict_of_vectors_50m_min, dict_of_vectors_5m_min = craters.ten_times_reduction(mg, vectors)
#
##This code snippet produces a smaller (800x800) grid where we let the basic crater routine run for a long time (1M impacts). The aim is to see if we get a steady state landscape.

#mg2 = RasterModelGrid()
#dict_of_vectors2_5m_min = {}
#mg2.initialize(600, 600, 0.0025)
#vectors2 = data(mg2)
#vectors2.elev[:] = 0.

#for i in xrange(20, 50):
#    print 'LOOP NUMBER ', i
#    mg2, vectors2, profile, xsec = craters.dig_some_craters(mg2, vectors2, nt_in=10000)
#    dict_of_vectors2_5m_min[i] = copy(vectors2)

#This code snippet to produce a landscape w. 200000 impacts, but where the first quarter of the impacts are 3x bigger on average. The output interval is only 5k impacts this time.
mg3 = RasterModelGrid()
mg3.initialize(1200, 1200, 0.0025)
vectors = data(mg3)
vectors.elev[:] = 0.

dict1_1200_50k_15m_min, dict1_1200_150k_5m_min = craters.step_reduce_size(mg3, vectors, loops=[10,30], interval=5000, min_radius_in=[0.015,0.005])

vectors.elev[:] = 0.

dict2_1200_150k_10m_min, dict2_1200_150k_5m_min = craters.step_reduce_size(mg3, vectors, loops=[30,30], interval=5000, min_radius_in=[0.01,0.005])





#! /usr/bin/env python

import unittest
import numpy as np
from numpy.testing import assert_array_equal

from landlab.field.grouped import ModelDataFields


class TestModelDataFields(unittest.TestCase):
    def test_init(self):
        fields = ModelDataFields()
        self.assertSetEqual(set(), fields.groups)

    def test_new_field_location(self):
        fields = ModelDataFields()
        fields.new_field_location('nodes', 12)
        self.assertSetEqual(set(['nodes']), fields.groups)

    def test_add_existing_group(self):
        fields = ModelDataFields()
        fields.new_field_location('nodes', 12)
        with self.assertRaises(ValueError):
            fields.new_field_location('nodes', 24)

    def test_add_multiple_groups(self):
        fields = ModelDataFields()
        fields.new_field_location('nodes', 12)
        fields.new_field_location('cells', 2)
        fields.new_field_location('faces', 7)
        fields.new_field_location('links', 7)
        self.assertSetEqual(set(['nodes', 'cells', 'faces', 'links']),
                            fields.groups)

    def test_ones(self):
        fields = ModelDataFields()
        fields.new_field_location('nodes', 12)
        fields.new_field_location('cells', 2)

        value_array = fields.ones('nodes')
        assert_array_equal(np.ones(12.), value_array)

        value_array = fields.ones('cells')
        assert_array_equal(np.ones(2.), value_array)

    def test_add_ones(self):
        fields = ModelDataFields()
        fields.new_field_location('nodes', 12)
        fields.new_field_location('cells', 2)

        fields.add_ones('nodes', 'z')
        assert_array_equal(np.ones(12.), fields['nodes']['z'])
        assert_array_equal(np.ones(12.), fields.field_values('nodes', 'z'))

        fields.add_ones('cells', 'z')
        assert_array_equal(np.ones(2.), fields['cells']['z'])
        assert_array_equal(np.ones(2.), fields.field_values('cells', 'z'))



#! /usr/bin/env python

import pylab

from landlab.components.flexure import Flexure

_SHAPE = (100, 100)
_SPACING = (10e3, 10e3)
_LOAD_LOCS = [
    (_SHAPE[0]/2, _SHAPE[1]/2),
    (_SHAPE[0]/4, _SHAPE[1]/4),
]


def main():

    flex = Flexure(_SHAPE, _SPACING, (0., 0.))

    load = flex['lithosphere__overlying_pressure']
    for loc in _LOAD_LOCS:
        load[loc] = 1e9

    dz = flex['lithosphere__elevation']

    flex.update()

    dz.shape = _SHAPE

    #pylab.imshow(dz)
    #pylab.colorbar()
    pylab.plot(dz[_SHAPE[0] * 3 / 8, :])
    pylab.show()


if __name__ == '__main__':
    main()



import numpy
from landlab import RasterModelGrid
from landlab.components.perron_nl_diff import perron_nl_diff_faster as perron_nl_diff
from landlab.components.craters import data
import pylab

class drive_perron(object):
    
    def __init__(self):
        self.nr = 100
        self.nc = 100
        self.dx = 1.
        self.dt = 1.
        self.nt = 10
        self.uplift = 1. #per tstep
        self.build_grid_and_data()
        self.diff = perron_nl_diff(self.mg, self.vectors, self.dt)
       
         
    def build_grid_and_data(self):
        '''
        Builds a nr*nc grid open at the top and bottom edges for testing
        '''
        self.mg = RasterModelGrid()
        self.mg.initialize(self.nr, self.nc, self.dx)
        self.mg.set_noflux_boundaries(False, True, False, True)  # Boundary conditions
        self.vectors = data()
        self.vectors.elev = [0.] * self.mg.ncells
    
    def update(self):
        for i in range(self.mg.ncells):
            self.vectors.elev[i] = self.vectors.elev[i] + self.uplift


def main():
    i = 0
    dp = drive_perron()
    while i<dp.nt:
        dp.update()
        dp.diff.update(dp.mg, dp.vectors)
        i = i+1
        print 'Loop ', i, ' complete'
        pylab.plot(range(dp.nr), dp.vectors.elev[1:(dp.mg.ncells+2-dp.nc):dp.nc])
    pylab.show()
    return dp
    

if __name__=='__main__':
    main()



#! /usr/bin/env python

import pylab
import numpy as np

from landlab.components.flexure import Flexure

#_SHAPE = (200, 200)
#_SPACING = (5e3, 5e3)
#_N_LOADS = 128


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--n-loads', type=int, default=1,
                        help='Number of loads to apply')
    parser.add_argument('--shape', type=int, default=128,
                        help='Number rows and columns')
    parser.add_argument('--spacing', type=int, default=5e3,
                        help='Spading between rows and columns (m)')
    parser.add_argument('--n-procs', type=int, default=1,
                        help='Number of processors to use')
    parser.add_argument('--plot', action='store_true', default=False,
                        help='Plot an image of the total deflection')

    args = parser.parse_args()

    shape = (args.shape, args.shape)
    spacing = (args.spacing, args.spacing)

    load_locs = np.random.random_integers(0, shape[0] * shape[1] - 1,
                                          args.n_loads)
    load_sizes = np.random.normal(1e9, 1e7, args.n_loads)

    flex = Flexure(shape, spacing, (0., 0.))
    load = flex['lithosphere__overlying_pressure']

    for (loc, size) in zip(load_locs, load_sizes):
        load.flat[loc] = size

    flex.update(n_procs=args.n_procs)

    if args.plot:
        pylab.imshow(flex['lithosphere__elevation'])
        pylab.colorbar()
        pylab.show()


if __name__ == '__main__':
    main()
